# built-in modules
from typing import List, Tuple, Generator, Union
from enum import IntEnum

# third-party modules
from ordered_set import OrderedSet

# qsynthesis modules
from qsynthesis.tables.base import InputOutputOracle
from qsynthesis.tritonast import TritonAst
from qsynthesis.algorithms.synthesizer_base import SynthesizerBase, logger


class YieldT(IntEnum):
    """
    Enum to identify the yield type in the huge
    generator that performs the AST search and
    substitution.
    """
    TopBottom = 1
    BottomUp = 2


class TopDownBottomUpSynthesizer(SynthesizerBase):
    """
    Synthesize with Top-Down then Bottom-Up AST search based on Triton AST.
    The idea behind this bi-directional search is that a node simplified in
    a sub-AST might enable synthesizing one of its parent by means of simplification
    if for instance we turn a two-variable AST in a one-variable AST.

    .. warning:: This class is not meant to be instanciated directly but rather
                 serving a common base for children class using this search strategy.
    """

    def __init__(self, ltms: Union[InputOutputOracle, List[InputOutputOracle]], only_first_match: bool = False, learning_new_exprs: bool = False):
        """
        Constructor that takes lookup tables as input.

        :param ltms: Single lookup table of a list of them
        :param only_first_match: boolean that stop interating over tables as soon as the lookup is successfull for one
        :param learning_new_exprs: boolean that enables improving the current table if if a synthesized entry appears
                                   to be bigger than the one submitted
        """
        super(TopDownBottomUpSynthesizer, self).__init__(ltms, only_first_match, learning_new_exprs)
        self.total_repl_td = 0
        self.total_repl_bu = 0

    def synthesize(self, ioast: TritonAst, check_sem: bool = False) -> Tuple[TritonAst, bool]:
        """
        Performs the Top-Down and then Bottom-Search to synthesize the AST

        :param ioast: TritonAst object to synthesize
        :param check_sem: boolean one whether to check the semantic equivalence of expression
                          before substituting them. That ensure soundness of the synthesis
        :returns: tuple with new TritonAst and whether some replacement took place or not

        .. warning:: Activating the `check_sem` parameter implies a strong overhead
                     on the synthesis as SMT queries are being performed for any candidates
        """
        ioast = ioast.duplicate()  # Make a copy of the ast not tamper it
        self.expr_cache = {}
        self.eval_cache = {}
        self.call_to_eval = 0
        self.eval_count = 0

        expr_repl_visitor = self._visit_replacement(ioast)
        new_expr_to_send = None
        expr_modified = False

        while 1:
            # logger.debug(f"Sending: {new_expr_to_send}")
            cur_expr, info = expr_repl_visitor.send(new_expr_to_send)  # Iterate generator
            # logger.debug(f"Receiving: {cur_expr}")
            if isinstance(info, bool):
                final_expr = cur_expr
                # logger.debug("final expression:", final_expr.replace("\n"," "))
                final_expr.update_all()
                return final_expr, expr_modified

            else:  # Try synthesizing expression
                logger.debug(f"try synthesis lookup: {cur_expr.pp_str if cur_expr.node_count < 50 else 'too large'}")
                synt_res = self.try_synthesis_lookup(cur_expr, check_sem)
                if synt_res is not None:
                    expr_modified = True
                new_expr_to_send = synt_res  # Send the result to the generator (thus either new expr or None)

    def _visit_replacement(self, ast: TritonAst, iff_replace=True, update=True) -> Generator[Tuple[TritonAst, Union[YieldT, bool]], TritonAst, None]:
        """
        Generator that will iterate over the AST. It will yield each sub-AST and is expecting
        through the send mechanism to receive None (meaning nothing has been found) or a TritonAst
        which means the AST just yielded have to be substituted by this one.

        .. note:: Deprecated to use. Use _visit_replacement_bfs instead
        """
        chs = ast.get_children()
        if not chs:  # If arity is 0 (variable and constant)
            yield ast, False  # Yield the final expr as-is
        else:
            # Normally here in Z3Ast check here if expression is bool
            rep = yield ast, YieldT.TopBottom  # First (Top-Down yield)

            if rep is not None:  # We should replace this expr
                self.total_repl_td += 1
                yield rep, True  # Final (yield)
            else:
                reps = {}  # index -> (expr, bool) the boolean true if expr was replaced
                for i, c in enumerate(chs):  # Iterate (and simplify) all childrens
                    g = self._visit_replacement(c, iff_replace, update)
                    recv = None
                    while 1:
                        it, info = g.send(recv)  # forward the new expression to the sub-generator
                        if isinstance(info, bool):  # is a final yield
                            reps[i] = it, info
                            break
                        else:
                            recv = yield it, info  # re-yield it above (and receive a new expression)

                if sum(x[1] for x in reps.values()) or not iff_replace:  # at least on replacement took place among children
                    for i, (e, b) in reps.items():
                        if b:  # expr has been replaced
                            ast.set_child(i, e)  # replace the child
                    if update:  # if one of the children got replaced and update, update fields
                        ast.update()
                    rep = yield ast, YieldT.BottomUp  # Second (Bottom-Up yield) (if not bool expr)
                    if rep is None:
                        yield ast, True  # Final yield
                    else:
                        self.total_repl_bu += 1
                        yield rep, True  # Final yield
                else:
                    yield ast, False

    def _visit_replacement_bfs(self, orig_ast: TritonAst) -> Generator[Tuple[TritonAst, Union[YieldT, bool]], TritonAst, None]:
        """
        Generator that will iterate over the AST. It will yield each sub-AST and is expecting
        through the send mechanism to receive None (meaning nothing has been found) or a TritonAst
        which means the AST just yielded have to be substituted by this one. The search is made in
        a Breath-First-Search manner as it provides better results for synthesis.

        :param orig_ast: TritonAst to iterate (and to modify)
        :returns: generator that always yield a TritonAst and either a yield type indicating whether it is during
                  the top-down or bottom-up search or a boolean indicating all substitutions have been performed
                  in the AST. The generate receive via send a new TritonAst or None
        """
        worklist = OrderedSet([orig_ast])
        bottomup_worklist = OrderedSet([orig_ast])
        modified = False
        mode = YieldT.TopBottom

        while worklist:
            ast = worklist.pop(0)
            logger.debug(f"\n[visit_replacement] pop [{ast.ptr_id}]({len(ast.get_children())}) {ast.pp_str}  worklist:[{len(worklist)}] bu_worklist:[{len(bottomup_worklist)}]")

            if not ast.is_leaf():
                rep = yield ast, mode  # Current yield (TopB, or BotUp)
            else:
                rep = None

            if rep is not None:  # We should replace this expr
                modified = True
                self.total_repl_td += 1 if mode == YieldT.TopBottom else 0
                self.total_repl_bu += 1 if mode == YieldT.BottomUp else 0
                if ast.is_root():
                    logger.debug(f"YIeld new root ahead of time: {rep.pp_str}")
                    yield rep, modified  # Final yield (ahead of time)
                else:
                    logger.debug(f"replace_self: {str(ast.expr)} => {rep.pp_str}")
                    ast.replace_self(rep, update_parents=True)
            else:
                if mode == YieldT.TopBottom:
                    logger.debug(f"[visit_replacement] worklist updates before wl:[{len(worklist)}] buwl:[{len(bottomup_worklist)}]")
                    worklist.update(ast.get_children())  # add child in the TopDown only if it has not been replaced
                    bottomup_worklist.add(ast)  # add itself in the BottomUp list !
                    logger.debug(f"[visit_replacement] worklist updates wl:[{len(worklist)}] buwl:[{len(bottomup_worklist)}]")

            if not worklist:
                logger.debug("--------------------------------- Switch Bottom Up -----------------------------------")
                logger.debug(f"AST at that time: {orig_ast.pp_str}")
                mode = YieldT.BottomUp
                worklist = bottomup_worklist[::-1]  # switch the worklist
                bottomup_worklist = []
        yield orig_ast, modified  # Final yield



# built-in imports
from typing import Set
from collections import deque, namedtuple

# third-party imports
from triton import AST_NODE

# local imports
from qsynthesis.types import Addr

NodeInfo = namedtuple("NodeInfo", "id expr_off address")


def backslice(root_sym_expr: 'SymbolicExpression') -> Set[Addr]:
    """
    Extract backslice of given symbolic expression as a DGraph
    of instructions ids.

    :param ctx: Triton context to extract the backslice from
    :param root_sym_expr: target symbolic expression
    :return: A DGraph where each node corresponds to an instruction id
    """

    addrs = set()

    # Add root node to graph
    root_node_id, off, addr = [int(x) for x in root_sym_expr.getComment().split('#')]
    addrs.add(addr)

    # Create a queue of (node, parent_id) tuples
    nodes_to_process = deque()
    nodes_to_process.append((root_sym_expr.getAst(), root_node_id))
    handled = set()

    while len(nodes_to_process) > 0:

        node, parent_id = nodes_to_process.popleft()
        skip_node = False

        # If the given AstNode is a reference consume it
        # and then continue using the AstNode of the referenced
        # symbolic expression
        while node.getType() == AST_NODE.REFERENCE:
            # Get referenced symbolic expr
            se = node.getSymbolicExpression()

            # Get expression id
            instr_id, expr_nb, instr_addr = [int(x) for x in se.getComment().split('#')]
            expr_id = (instr_id, expr_nb)

            addrs.add(instr_addr)

            # Do not process this node if it has already been handled
            # NOTE: since we could come from an unseen parent we add
            # an edge before going to the next node
            if expr_id in handled:
                skip_node = True
                break
            else:
                handled.add(expr_id)

            # Move to the referenced AstNode
            parent_id = instr_id
            node = node.getSymbolicExpression().getAst()

        if skip_node:
            continue

        # Process children
        for child in node.getChildren():
            nodes_to_process.append((child, parent_id))

    return addrs



# built-in libs
from __future__ import annotations
from pathlib import Path
from enum import IntEnum
import array
import hashlib
import threading
from collections import Counter
from time import time, sleep
import ctypes
import logging

# third-party libs
import psutil

# qsynthesis deps
from qsynthesis.grammar import TritonGrammar
from qsynthesis.tritonast import TritonAst
from qsynthesis.types import AstNode, Hash, Optional, List, Dict, Union, Tuple, Iterable, Input, Output, BitSize, Any, \
                             Generator

logger = logging.getLogger("qsynthesis")


class HashType(IntEnum):
    """
    Hash types supported by the Lookup table database. In practice solely md5
    is used, has it is the fastest of all
    """
    RAW = 1
    FNV1A_128 = 2
    MD5 = 3


class _EvalCtx(object):
    """
    Small debugging Triton evaluation context. It is used when manipulating
    tables in a standalone manner. It enables obtaining TritonAst out of
    the databqse entries.
    """
    def __init__(self, grammar):
        from triton import TritonContext, ARCH, AST_REPRESENTATION
        # Create the context
        self.ctx = TritonContext(ARCH.X86_64)
        self.ctx.setAstRepresentationMode(AST_REPRESENTATION.PYTHON)
        self.ast = self.ctx.getAstContext()

        # Create symbolic variables for grammar variables
        self.symvars = {}
        self.vars = {}
        for v, sz in grammar.vars_dict.items():
            sym_v = self.ctx.newSymbolicVariable(sz, v)
            self.symvars[v] = sym_v
            self.vars[v] = self.ast.variable(sym_v)

        # Create mapping to triton operators
        self.tops = {x: getattr(self.ast, x) for x in dir(self.ast) if not x.startswith("__")}

    def eval_str(self, s: str) -> AstNode:
        """Eval the string expression to create an AstNode object"""
        e = eval(s, self.tops, self.vars)
        bv_size = list(self.vars.values())[0].getBitvectorSize()  # Assume all vars are of same size
        if isinstance(e, int):  # In case the expression was in fact an int
            return self.ast.bv(e, bv_size)
        else:
            return e

    def set_symvar_values(self, args: Input) -> None:
        for v_name, value in args.items():
            self.ctx.setConcreteVariableValue(self.symvars[v_name], value)


class InputOutputOracle:
    """
    Base Lookup table class. Specify the interface that child class have to
    implement to be interoperable with other the synthesizer.
    """
    def __init__(self, gr: TritonGrammar, inputs: List[Input], hash_mode: HashType = HashType.RAW, f_name: Union[Path, str] = ""):
        """
        Constructor making a I/O oracle from a grammar a set of inputs and an hash type.

        :param gr: triton grammar
        :param inputs: List of inputs
        :param hash_mode: type of hash to be used as keys in tables
        :param f_name: file name of the table (when being loaded)
        """
        self._name = Path(f_name)
        self.grammar = gr
        self._bitsize = self.grammar.size
        self.expr_cache = {}
        self.lookup_count = 0
        self.lookup_found = 0
        self.cache_hit = 0
        self.hash_mode = hash_mode
        self._ectx = None
        # generation related fields
        self.watchdog = None
        self.max_mem = 0
        self.stop = False

        self.inputs = inputs

    @property
    def size(self) -> int:
        """Size of the table (number of entries)

        :rtype: int
        """
        raise NotImplementedError("Should be implemented by child class")

    def _get_item(self, h: Hash) -> Optional[str]:
        """
        From a given hash return the associated expression string if
        found in the lookup table.

        :param h: hash of the item to get
        :returns: raw expression string if found
        """
        raise NotImplementedError("Should be implemented by child class")

    def is_expr_compatible(self, expr: TritonAst) -> bool:
        """
        Check the compatibility of the given expression with the table.
        The function checks sizes of expr variables against the one of
        its own grammar.

        :param expr: TritonAst expression to check
        :return: True if the table can decide on this expression
        """
        e_vars = Counter(x.getBitSize() for x in expr.symvars)
        e_table = Counter(self.grammar.vars_dict.values())
        for sz, count in e_vars.items():
            if sz in e_table:
                if count > e_table[sz]:
                    return False
            else:
                return False
        return True

    def lookup(self, outputs: List[Output], *args,  use_cache: bool = True) -> Optional[TritonAst]:
        """
        Perform a lookup in the table with a given set of outputs corresponding
        to the evaluation of an AST against the Input of this exact same table.
        If an entry is found a TritonAst is created and returned.

        :param outputs: list of output result of evaluating an ast against the inputs of this table
        :type: List[:py:obj:`qsynthesis.types.Output`]
        :param args: args forwarded to grammar and ultimately to the tritonAst in charge of building a new TritonAst
        :param use_cache: Boolean enabling caching the the hash of outputs. A second call if the same outputs
                          (which is common) will not trigger a lookup in the database
        :returns: optional TritonAst corresponding of the expression found in the table
        """
        self.lookup_count += 1
        h = self.hash(outputs)
        if h in self.expr_cache and use_cache:
            self.cache_hit += 1
            return self.expr_cache[h]
        else:
            v = self._get_item(h)
            if v:
                self.lookup_found += 1
                try:
                    e = self.grammar.str_to_expr(v, *args)
                    self.expr_cache[h] = e
                    return e
                except NameError:
                    return None
                except TypeError:
                    return None
            else:
                return None

    def lookup_hash(self, h: Hash) -> Optional[str]:
        """
        Raw lookup for a given key in database.

        :param h: hash key to look for in database
        :type h: :py:obj:`qsynthesis.types.Hash`
        :returns: string of the expression if found
        :rtype: Optional[str]
        """
        return self._get_item(h)

    @property
    def is_writable(self) -> bool:
        """ Whether the table enable being written (with new expressions)

        :rtype: bool
        """
        return False

    @property
    def name(self) -> str:
        """ Name of the table

        :rtype: str
        """
        return str(self._name)

    @property
    def bitsize(self) -> BitSize:
        """ Size of expression in bit

        :rtype: :py:obj:`qsynthesis.types.BitSize`
        """
        return self._bitsize

    @property
    def var_number(self) -> int:
        """ Maximum number of variables contained in the table

        :rtype: int
        """
        return len(self.grammar.vars)

    @property
    def operator_number(self) -> int:
        """ Number of operators used in this table

        :rtype: int
        """
        return len(self.grammar.ops)

    @property
    def input_number(self) -> int:
        """ Number of inputs used in this table

        :rtype: int
        """
        return len(self.inputs)

    @staticmethod
    def _fnv1a_128(outs: List[Output]) -> Hash:
        """
        Hash the outputs using fnv1a_128 algorithm

        :param outs: list of outputs to hash
        :type outs: List[:py:obj:`qsynthesis.types.Output`]
        :returns: Hash value (int) corresponding to the fnv1a of outputs
        :rtype: :py:obj:`qsynthesis.types.Hash`
        """
        a = array.array('Q', outs)
        FNV1A_128_OFFSET = 0x6c62272e07bb014262b821756295c58d
        FNV1A_128_PRIME = 0x1000000000000000000013b  # 2^88 + 2^8 + 0x3b

        # Set the offset basis
        hash = FNV1A_128_OFFSET

        # For each character
        for byte in a.tobytes():
            # Xor with the current character
            hash ^= byte
            # Multiply by prime
            hash *= FNV1A_128_PRIME
            # Clamp
            hash &= 0xffffffffffffffffffffffffffffffff
        # Return the final hash as a number
        return hash

    @staticmethod
    def _md5(outs: List[Output]) -> Hash:
        """
        Hash the outputs using MD5 algorithm. Outputs are transformed into an array.
        That means the final bytes hashed are the concatenation of uint64 in little
        endian.

        :param outs: list of outputs to hash
        :type outs: List[:py:obj:`qsynthesis.types.Output`]
        :returns: Bytes corresponding to MD5 hash
        :type: :py:obj:`qsynthesis.types.Hash`
        """
        a = array.array('Q', outs)
        h = hashlib.md5(a.tobytes())
        return h.digest()

    def hash(self, outs: List[Output]) -> Hash:
        """
        Main hashing method that dispatch the outputs to the appropriate hashing
        function depending on the ``hash_mode`` of the table.

        :param outs: list of outputs to hash
        :type outs: List[:py:obj:`qsynthesis.types.Output`]
        :returns: Hash type (bytes, int ..) of the outputs
        :rtype: :py:obj:`qsynthesis.types.Hash`
        """
        if self.hash_mode == HashType.RAW:
            return tuple(outs)
        elif self.hash_mode == HashType.FNV1A_128:
            return self._fnv1a_128(outs)
        elif self.hash_mode == HashType.MD5:
            return self._md5(outs)

    def __iter__(self) -> Iterable[Tuple[Hash, str]]:
        """ Iterator of all the entries as an iterator of pair, hash, expression as string

        :rtype: Iterable[Tuple[:py:obj:`qsynthesis.types.Hash, str]]`
        """
        raise NotImplementedError("Should be implemented by child class")

    def _get_expr(self, expr: str) -> AstNode:
        """
        Utility function that returns a TritonAst from a given expression string.
        A TritonContext local to the table is created to enable generating such ASTs.

        :param expr: Expression
        :returns: TritonAst resulting of the parsing of s
        """
        if self._ectx is None:
            self._ectx = _EvalCtx(self.grammar)
        return self._ectx.eval_str(expr)

    def _set_input_lcontext(self, i: Union[int, Input]) -> None:
        """
        Set the given concrete values of variables in the local TritonContext.
        The parameter is either the ith input of the table, or directly an Input
        given a valuation for each variables. This function must be called before
        performing any evaluation of an AST.

        :param i: index of the input, or Input object (dict)
        :returns: None
        """
        if self._ectx is None:
            self._ectx = _EvalCtx(self.grammar)
        self._ectx.set_symvar_values(self.inputs[i] if isinstance(i, int) else i)

    def _eval_expr_inputs(self, expr: AstNode) -> List[Output]:
        """
        Evaluate a given Triton AstNode object on all inputs of the
        table. The result is a list of Output values.

        :param expr: Triton AstNode to evaluate
        :type expr: :py:obj:`qsynthesis.types.AstNode`
        :returns: list of output values (ready to be hashed)
        :rtype: List[:py:obj:`qsynthesis.types.Output`]
        """
        outs = []
        for i in range(len(self.inputs)):
            self._set_input_lcontext(i)
            outs.append(expr.evaluate())
        return outs

    def _watchdog_worker(self, threshold: Union[float, int]) -> None:
        """
        Function where the memory watchdog thread is running. This function
        allows interrupting table generation when it happens to fill the
        given threshold of RAM.

        :param threshold: percentage of RAM load that triggers the stop of generation
        """
        while not self.stop:
            sleep(2)
            mem = psutil.virtual_memory()
            self.max_mem = max(mem.used, self.max_mem)
            if mem.percent >= threshold:
                logger.warning(f"Threshold reached: {mem.percent}%")
                self.stop = True  # Should stop self and also main thread

    @staticmethod
    def _try_linearize(s: str, symbols: Dict[str, object]) -> str:
        """
        Try applying sympy to linearize ``s`` with the variable symbols
        ``symbols``. If any exception is raised in between to expression
        string is returned unchanged.

        :param s: expression string to linearize
        :param symbols: dictionnary of variables names to sympy symbol objects

        .. warning:: This function requires sympy to be installed !
        """
        import sympy
        try:
            lin = eval(s, symbols)
            if isinstance(lin, sympy.boolalg.BooleanFalse):
                logger.error(f"[linearization] expression {s} False")
            logger.debug(f"[linearization] expression linearized {s} => {lin}")
            return str(lin).replace(" ", "")
        except TypeError:
            return s
        except AttributeError as _:
            return s

    @staticmethod
    def _to_signed(value: int) -> int:
        return ctypes.c_longlong(value).value

    @staticmethod
    def _to_unsigned(value: int) -> int:
        return ctypes.c_ulonglong(value).value

    @staticmethod
    def _is_constant(v1: str) -> bool:
        try:
            int(v1)
            return True
        except ValueError:
            return False

    @staticmethod
    def _custom_permutations(l: List[Any]) -> Generator[Tuple[bool, Any, Any], None, None]:
        """
        Custom generator generating all the possible tuples from a list. But instead
        of iterating item i with all others 0..n, iterates i with all the previous 0..i.
        It generates a somewhat sorted generated that ensure pairs of items appearing
        first in the list will be yielded before.

        :param l: list of any item
        :returns: genreator of tuples generating all possibles pairs
        """
        for i in range(len(l)):
            for j in range(0, i):
                yield False, l[i], l[j]
                yield False, l[j], l[i]
            yield True, l[i], l[i]

    def generate(self,
                 bitsize: int,
                 constants: List[int] = [],
                 do_watch: bool = False,
                 watchdog_threshold: Union[int, float] = 90,
                 linearize: bool = False,
                 do_use_blacklist: bool = False,
                 limit: int = 0) -> None:
        """
        Generate a new lookup table from scratch with the variables and operators
        set in the constructor of the table.

        :param bitsize: Bitsize of expressions to generate
        :param constants: List of constants to use in the generation
        :param do_watch: Enable RAM watching thread to monitor memory
        :param watchdog_threshold: threshold to be sent to the memory watchdog
        :param linearize: whether or not to apply linearization on expressions
        :param do_use_blacklist: enable blacklist mechanism on commutative operators. Slower but less memory consuming
        :param limit: Maximum number of entries to generate
        :returns: None
        """
        if do_watch:
            self.watchdog = threading.Thread(target=self._watchdog_worker, args=[watchdog_threshold], daemon=True)
            logger.debug("Start watchdog")
            self.watchdog.start()
        if linearize:
            logger.info("Linearization enabled")
            import sympy
            symbols = {x: sympy.symbols(x) for x in self.grammar.vars}
        t0 = time()

        from qsynthesis.grammar import jitting  # Import it locally to make sure pydffi is not mandatory
        CU = jitting.make_compilation_unit(bitsize)
        N = self.input_number
        ArTy = jitting.get_native_array_type(bitsize, N)

        hash_fun = lambda x: hashlib.md5(bytes(x)).digest() if self.hash_mode == HashType.MD5 else self.hash

        # Initialize worklist with variables
        worklist = [(ArTy(), k) for k in self.grammar.vars]
        for i, inp in enumerate(self.inputs):
            for v, k in worklist:
                v[i] = inp[k]

        # Initialize worklist with constants
        csts = [(ArTy(), str(c)) for c in constants]
        for (ar, c) in csts:
            jitting.init_array_cst(ar, int(c), N, bitsize)
        worklist.extend(csts)

        # initialize set of hash
        hash_set = set(hash_fun(x[0]) for x in worklist)

        ops = sorted(self.grammar.non_terminal_operators, key=lambda x: x.arity == 1)  # sort operators to iterate on unary first
        cur_depth = 2
        blacklist = set()
        item_count = len(worklist)  # total number of expressions

        try:
            while cur_depth > 0:
                # Start a new depth
                n_items = len(worklist)  # number of items to process at a given depth
                t = time() - t0
                print(f"Depth {cur_depth} (size:{n_items}) (Time:{int(t/60)}m{t%60:.5f}s)")
                c = 0

                for i, (same, (vals1, name1), (vals2, name2)) in enumerate(self._custom_permutations(worklist)):
                    if same:
                        c += 1
                        print(f"process: {(c*100)/n_items:.2f}%\r", end="")

                    if 0 < limit <= item_count:
                        self.stop = True

                    if self.stop:
                        logger.warning("Threshold reached, generation interrupted")
                        raise KeyboardInterrupt()

                    # Check it here once then iterate operators
                    name1_cst, name2_cst = self._is_constant(name1), self._is_constant(name2)
                    is_both_constant = name1_cst & name2_cst

                    for op, op_eval in zip(ops, [jitting.get_op_eval_array(CU, x) for x in ops]):  # Iterate over all operators
                        if op.arity == 1:
                            new_vals = ArTy()
                            op_eval(new_vals, vals1, N)

                            h = hash_fun(new_vals)
                            if h not in hash_set:
                                if name1_cst:
                                    fmt = str(self._to_signed(new_vals[0]))  # any value is the new constant value
                                else:
                                    fmt = f"{op.symbol}({name1})" if len(name1) > 1 else f"{op.symbol}{name1}"
                                fmt = self._try_linearize(fmt, symbols) if linearize else fmt
                                logger.debug(f"[add] {fmt: <20} {h}")
                                hash_set.add(h)
                                item_count += 1
                                worklist.append((new_vals, fmt))  # add it in worklist if not already in LUT
                            else:
                                logger.debug(f"[drop] {op.symbol}{name1}  ")

                        else:  # arity is 2
                            # for identity (a op a) ignore it if the result is known to be 0 or a
                            if same and (op.id_eq or op.id_zero):
                                continue

                            sn1 = f'{name1}' if len(name1) == 1 else f'({name1})'
                            sn2 = f'{name2}' if len(name2) == 1 else f'({name2})'
                            fmt = f"{op.symbol}({name1},{name2})" if op.is_prefix else f"{sn1}{op.symbol}{sn2}"

                            if not linearize:
                                if fmt in blacklist:  # Ignore expression if they are in the blacklist
                                    continue

                            new_vals = ArTy()

                            op_eval(new_vals, vals1, vals2, N)

                            if is_both_constant:  # if both were constant use the constant as repr instead
                                fmt = str(self._to_signed(new_vals[0]))

                            h = hash_fun(new_vals)
                            if h not in hash_set:
                                if linearize:
                                    fmt = self._try_linearize(fmt, symbols) if linearize else fmt
                                    if fmt in blacklist:  # if linearize check blacklist here
                                        continue

                                logger.debug(f"[add] {fmt: <20} {h}")
                                hash_set.add(h)
                                item_count += 1
                                worklist.append((new_vals, fmt))

                                if op.commutative and do_use_blacklist and not is_both_constant:
                                    fmt = f"{op.symbol}({name2},{name1})" if op.is_prefix else f"{sn2}{op.symbol}{sn1}"
                                    fmt = self._try_linearize(fmt, symbols) if linearize else fmt
                                    blacklist.add(fmt)  # blacklist commutative equivalent e.g for a+b blacklist: b+a
                                    logger.debug(f"[blacklist] {fmt}")
                            else:
                                logger.debug(f"[drop] {op.symbol}({name1},{name2})" if op.is_prefix else f"[drop] ({name1}){op.symbol}({name2})")

                cur_depth += 1
        except KeyboardInterrupt:
            logger.info("Stop required")
        # In the end
        self.stop = True
        t = time() - t0
        print(f"Depth {cur_depth} (size:{len(worklist)}) (Time:{int(t/60)}m{t%60:.5f}s) [RAM:{self.__size_to_str(self.max_mem)}]")
        self.add_entries(worklist, calc_hash=True)
        if do_watch:
            self.watchdog.join()

    def add_entry(self, hash: Hash, value: str) -> None:
        """
        Abstract function to add an entry in the lookuptable.

        :param hash: already computed hash to add
        :type Hash: :py:obj:`qsynthesis.types.Hash`
        :param value: expression value to add in the table
        :type value: str
        """
        raise NotImplementedError("Should be implemented by child class")

    def add_entries(self, worklist: List[Tuple[Hash, str]], calc_hash: bool = False) -> None:
        """
        Add the given list of entries in the database. The boolean ``calc_hash`` indicates
        whether hashes are already computed or not. If false the function should hash the
        hash first.

        :param worklist: list of entries to add
        :type worklist: List[Tuple[:py:obj:`qsynthesis.types.Hash`, str]]
        :param calc_hash: whether or not hash should be performed on entries keys
        :returns: None
        """
        raise NotImplementedError("Should be implemented by child class")

    @staticmethod
    def create(filename: Union[str, Path], grammar: TritonGrammar, inputs: List[Input], hash_mode: HashType = HashType.RAW, constants: List[int] = []) -> 'InputOutputOracle':
        """
        Create a new empty lookup table with the given initial parameters, grammars, inputs
        and hash_mode.

        :param filename: filename of the table to create
        :param grammar: TritonGrammar object representing variables and operators
        :param inputs: list of inputs on which to perform evaluation
        :type inputs: List[:py:obj:`qsynthesis.types.Input`]
        :param hash_mode: Hashing mode for keys
        :param constants: list of constants used
        :returns: lookuptable instance object
        """
        raise NotImplementedError("Should be implemented by child class")

    @staticmethod
    def load(file: Union[Path, str]) -> 'InputOutputOracle':
        """
        Load the given lookup table and returns an instance object.

        :param file: Database file to load
        :returns: InputOutputOracle object
        """
        raise NotImplementedError("Should be implemented by child class")

    @staticmethod
    def __size_to_str(value: int) -> str:
        """ Return pretty printed representation of RAM usage for table generation """
        units = [(float(1024), "Kb"), (float(1024 ** 2), "Mb"), (float(1024 ** 3), "Gb")]
        for unit, s in units[::-1]:
            if value / unit < 1:
                continue
            else:  # We are on the right unit
                return f"{value/unit:.2f}{s}"
        return f"{value}B"



# Standard modules
from __future__ import annotations
import random

# Qsynthesis types
from qsynthesis.types import BitSize, Char, Input, Dict, List, Tuple
from qsynthesis.tritonast import TritonAst
from qsynthesis.grammar.ops import BvOp, Operator, OPERATORS


class TritonGrammar(object):
    """
    Triton Grammar class. It represent a set of operators, and variables
    of a given size (only 64 bits at the moment).
    """

    def __init__(self, vars: List[Tuple[Char, BitSize]], ops: List[BvOp]):
        """
        Constructor taking a set of variables (name and size) and a set of operators.

        :param vars: list of tuple of (name ,size)
        :type vars: List[Tuple[:py:obj:`qsynthesis.types.Char`, :py:obj:`qsynthesis.types.BitSize`]]
        :param ops: list of BvOp representing operators
        :type ops: List[BvOp]
        """
        self.ops = ops
        self.vars_dict = {x[0]: x[1] for x in vars}  # Dict of str->size
        self.vars = list(self.vars_dict.keys())

        self.size = self.vars_dict[self.vars[0]]  # take size of the first var as they all have the same size

    @property
    def non_terminal_operators(self) -> List[Operator]:
        """
        Return the list of non-terminal operators. All unary and
        binary operators are non terminal as they can be derived.

        :return: list of operators namedtuples
        """
        return [OPERATORS[x] for x in self.ops]

    def gen_test_inputs(self, n: int) -> List[Input]:
        """
        Generate a list of ``n`` input. Thus it generate a random
        valuation for each variables of the grammar and that n times.

        :param n: Number of Input to generate (size of the list)
        :type n: int
        :returns: list of inputs
        :rtype: List[:py:obj:`qsynthesis.types.Input`]
        """
        return [{var: random.getrandbits(self.vars_dict[var]) for var in self.vars} for _ in range(n)]

    def str_to_expr(self, s: str, *args) -> TritonAst:
        """
        Convert a string in the format of the grammar into a TritonAst.
        In practice an args[0] should be a TritonAst from which to spawn
        a new TritonAst. That is required to get the same mapping of normalized
        variables than the one used by expr.

        :param s: expression string to convert to TritonAst
        :return: the TritonAst representing the expressions string
        :raises: NameError, TypeError
        """
        expr = args[0]
        return expr.normalized_str_to_ast(s)

    def to_dict(self) -> Dict:
        """
        Return a dictionnary representation of the grammar.
        This is used for serialization in database etc.
        """
        return dict(
            vars=[(n, sz) for n, sz in self.vars_dict.items()],
            operators=[x.name for x in self.ops]
        )

    @staticmethod
    def from_dict(g_dict: Dict) -> 'TritonGrammar':
        """
        Static method instanciating a TritonGrammar from its representation as
        a dictionnary.

        :param g_dict: dictionarry representation of the grammar
        :returns: TritonGrammar object
        """
        return TritonGrammar(g_dict['vars'], [BvOp[x] for x in g_dict['operators']])



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'synthesis_view.ui'
#
# Created by: PyQt5 UI code generator 5.14.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_synthesis_view(object):
    def setupUi(self, synthesis_view):
        synthesis_view.setObjectName("synthesis_view")
        synthesis_view.resize(523, 622)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(synthesis_view)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QtWidgets.QScrollArea(synthesis_view)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 558, 584))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.main_title = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.main_title.sizePolicy().hasHeightForWidth())
        self.main_title.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.main_title.setFont(font)
        self.main_title.setScaledContents(False)
        self.main_title.setAlignment(QtCore.Qt.AlignCenter)
        self.main_title.setObjectName("main_title")
        self.verticalLayout_3.addWidget(self.main_title)
        self.configurationBox = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.configurationBox.setObjectName("configurationBox")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.configurationBox)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.configurationVLayout = QtWidgets.QVBoxLayout()
        self.configurationVLayout.setObjectName("configurationVLayout")
        self.traceLayout = QtWidgets.QHBoxLayout()
        self.traceLayout.setObjectName("traceLayout")
        self.trace_label = QtWidgets.QLabel(self.configurationBox)
        self.trace_label.setObjectName("trace_label")
        self.traceLayout.addWidget(self.trace_label)
        self.trace_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.trace_type_box.setObjectName("trace_type_box")
        self.traceLayout.addWidget(self.trace_type_box)
        self.trace_line = QtWidgets.QLineEdit(self.configurationBox)
        self.trace_line.setObjectName("trace_line")
        self.traceLayout.addWidget(self.trace_line)
        self.configurationVLayout.addLayout(self.traceLayout)
        self.targetLayout = QtWidgets.QHBoxLayout()
        self.targetLayout.setObjectName("targetLayout")
        self.from_label = QtWidgets.QLabel(self.configurationBox)
        self.from_label.setEnabled(False)
        self.from_label.setObjectName("from_label")
        self.targetLayout.addWidget(self.from_label)
        self.from_line = QtWidgets.QLineEdit(self.configurationBox)
        self.from_line.setEnabled(False)
        self.from_line.setPlaceholderText("")
        self.from_line.setObjectName("from_line")
        self.targetLayout.addWidget(self.from_line)
        self.to_label = QtWidgets.QLabel(self.configurationBox)
        self.to_label.setEnabled(False)
        self.to_label.setObjectName("to_label")
        self.targetLayout.addWidget(self.to_label)
        self.to_line = QtWidgets.QLineEdit(self.configurationBox)
        self.to_line.setEnabled(False)
        self.to_line.setObjectName("to_line")
        self.targetLayout.addWidget(self.to_line)
        self.target_label = QtWidgets.QLabel(self.configurationBox)
        self.target_label.setEnabled(False)
        self.target_label.setObjectName("target_label")
        self.targetLayout.addWidget(self.target_label)
        self.target_box = QtWidgets.QComboBox(self.configurationBox)
        self.target_box.setEnabled(False)
        self.target_box.setObjectName("target_box")
        self.targetLayout.addWidget(self.target_box)
        self.register_box = QtWidgets.QComboBox(self.configurationBox)
        self.register_box.setEnabled(False)
        self.register_box.setObjectName("register_box")
        self.targetLayout.addWidget(self.register_box)
        self.mem_line = QtWidgets.QLineEdit(self.configurationBox)
        self.mem_line.setEnabled(False)
        self.mem_line.setObjectName("mem_line")
        self.targetLayout.addWidget(self.mem_line)
        self.operand_label = QtWidgets.QLabel(self.configurationBox)
        self.operand_label.setEnabled(False)
        self.operand_label.setObjectName("operand_label")
        self.targetLayout.addWidget(self.operand_label)
        self.configurationVLayout.addLayout(self.targetLayout)
        self.tableLayout = QtWidgets.QHBoxLayout()
        self.tableLayout.setObjectName("tableLayout")
        self.table_label = QtWidgets.QLabel(self.configurationBox)
        self.table_label.setEnabled(False)
        self.table_label.setObjectName("table_label")
        self.tableLayout.addWidget(self.table_label)
        self.table_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.table_type_box.setEnabled(False)
        self.table_type_box.setObjectName("table_type_box")
        self.tableLayout.addWidget(self.table_type_box)
        self.table_line = QtWidgets.QLineEdit(self.configurationBox)
        self.table_line.setEnabled(False)
        self.table_line.setObjectName("table_line")
        self.tableLayout.addWidget(self.table_line)
        self.configurationVLayout.addLayout(self.tableLayout)
        self.algorithmLayout = QtWidgets.QHBoxLayout()
        self.algorithmLayout.setObjectName("algorithmLayout")
        self.algorithm_label = QtWidgets.QLabel(self.configurationBox)
        self.algorithm_label.setEnabled(False)
        self.algorithm_label.setObjectName("algorithm_label")
        self.algorithmLayout.addWidget(self.algorithm_label)
        self.algorithm_box = QtWidgets.QComboBox(self.configurationBox)
        self.algorithm_box.setEnabled(False)
        self.algorithm_box.setObjectName("algorithm_box")
        self.algorithmLayout.addWidget(self.algorithm_box)
        self.algorithm_type_label = QtWidgets.QLabel(self.configurationBox)
        self.algorithm_type_label.setEnabled(False)
        self.algorithm_type_label.setObjectName("algorithm_type_label")
        self.algorithmLayout.addWidget(self.algorithm_type_label)
        self.algorithm_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.algorithm_type_box.setEnabled(False)
        self.algorithm_type_box.setObjectName("algorithm_type_box")
        self.algorithmLayout.addWidget(self.algorithm_type_box)
        self.qtrace_sym_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.qtrace_sym_type_box.setEnabled(False)
        self.qtrace_sym_type_box.setObjectName("qtrace_sym_type_box")
        self.algorithmLayout.addWidget(self.qtrace_sym_type_box)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.algorithmLayout.addItem(spacerItem)
        self.configurationVLayout.addLayout(self.algorithmLayout)
        self.experimentalLayout = QtWidgets.QHBoxLayout()
        self.experimentalLayout.setObjectName("experimentalLayout")
        self.label_9 = QtWidgets.QLabel(self.configurationBox)
        self.label_9.setObjectName("label_9")
        self.experimentalLayout.addWidget(self.label_9)
        self.verticalLayout_9 = QtWidgets.QVBoxLayout()
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.learning_checkbox = QtWidgets.QCheckBox(self.configurationBox)
        self.learning_checkbox.setObjectName("learning_checkbox")
        self.verticalLayout_9.addWidget(self.learning_checkbox)
        self.check_sem_checkbox = QtWidgets.QCheckBox(self.configurationBox)
        self.check_sem_checkbox.setObjectName("check_sem_checkbox")
        self.verticalLayout_9.addWidget(self.check_sem_checkbox)
        self.experimentalLayout.addLayout(self.verticalLayout_9)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.experimentalLayout.addItem(spacerItem1)
        self.configurationVLayout.addLayout(self.experimentalLayout)
        self.verticalLayout_6.addLayout(self.configurationVLayout)
        self.verticalLayout_3.addWidget(self.configurationBox)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.tritonLayout = QtWidgets.QVBoxLayout()
        self.tritonLayout.setObjectName("tritonLayout")
        self.run_triton_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.run_triton_button.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.run_triton_button.sizePolicy().hasHeightForWidth())
        self.run_triton_button.setSizePolicy(sizePolicy)
        self.run_triton_button.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.run_triton_button.setFlat(False)
        self.run_triton_button.setObjectName("run_triton_button")
        self.tritonLayout.addWidget(self.run_triton_button, 0, QtCore.Qt.AlignHCenter)
        self.triton_textarea = QtWidgets.QTextBrowser(self.scrollAreaWidgetContents)
        self.triton_textarea.setEnabled(False)
        self.triton_textarea.setObjectName("triton_textarea")
        self.tritonLayout.addWidget(self.triton_textarea)
        self.triton_button_group = QtWidgets.QHBoxLayout()
        self.triton_button_group.setObjectName("triton_button_group")
        self.show_deps_triton_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.show_deps_triton_button.setEnabled(False)
        self.show_deps_triton_button.setObjectName("show_deps_triton_button")
        self.triton_button_group.addWidget(self.show_deps_triton_button)
        self.line_2 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.triton_button_group.addWidget(self.line_2)
        self.show_ast_triton_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.show_ast_triton_button.setEnabled(False)
        self.show_ast_triton_button.setObjectName("show_ast_triton_button")
        self.triton_button_group.addWidget(self.show_ast_triton_button)
        self.tritonLayout.addLayout(self.triton_button_group)
        self.horizontalLayout_2.addLayout(self.tritonLayout)
        self.line = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout_2.addWidget(self.line)
        self.synthesisLayout = QtWidgets.QVBoxLayout()
        self.synthesisLayout.setObjectName("synthesisLayout")
        self.run_synthesis_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.run_synthesis_button.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.run_synthesis_button.sizePolicy().hasHeightForWidth())
        self.run_synthesis_button.setSizePolicy(sizePolicy)
        self.run_synthesis_button.setObjectName("run_synthesis_button")
        self.synthesisLayout.addWidget(self.run_synthesis_button, 0, QtCore.Qt.AlignHCenter)
        self.synthesis_textarea = QtWidgets.QTextBrowser(self.scrollAreaWidgetContents)
        self.synthesis_textarea.setEnabled(False)
        self.synthesis_textarea.setObjectName("synthesis_textarea")
        self.synthesisLayout.addWidget(self.synthesis_textarea)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.show_ast_synthesis_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.show_ast_synthesis_button.setEnabled(False)
        self.show_ast_synthesis_button.setObjectName("show_ast_synthesis_button")
        self.horizontalLayout_4.addWidget(self.show_ast_synthesis_button)
        self.line_3 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line_3.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.horizontalLayout_4.addWidget(self.line_3)
        self.reassemble_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.reassemble_button.setEnabled(False)
        self.reassemble_button.setObjectName("reassemble_button")
        self.horizontalLayout_4.addWidget(self.reassemble_button)
        self.synthesisLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_2.addLayout(self.synthesisLayout)
        self.verticalLayout_3.addLayout(self.horizontalLayout_2)
        self.verticalLayout_4.addLayout(self.verticalLayout_3)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(synthesis_view)
        QtCore.QMetaObject.connectSlotsByName(synthesis_view)

    def retranslateUi(self, synthesis_view):
        _translate = QtCore.QCoreApplication.translate
        synthesis_view.setWindowTitle(_translate("synthesis_view", "QSynthesis"))
        self.main_title.setText(_translate("synthesis_view", "QSynthesis"))
        self.configurationBox.setTitle(_translate("synthesis_view", "Synthesis configuration"))
        self.trace_label.setText(_translate("synthesis_view", "Trace:"))
        self.from_label.setText(_translate("synthesis_view", "From:"))
        self.to_label.setText(_translate("synthesis_view", "To:"))
        self.target_label.setText(_translate("synthesis_view", "Target:"))
        self.operand_label.setText(_translate("synthesis_view", "TextLabel"))
        self.table_label.setText(_translate("synthesis_view", "Table:"))
        self.algorithm_label.setText(_translate("synthesis_view", "Algorithm:"))
        self.algorithm_type_label.setText(_translate("synthesis_view", "Type:"))
        self.label_9.setText(_translate("synthesis_view", "Experimental settings:"))
        self.learning_checkbox.setText(_translate("synthesis_view", "Learning new expressions"))
        self.check_sem_checkbox.setText(_translate("synthesis_view", "Check semantic expressions"))
        self.run_triton_button.setText(_translate("synthesis_view", "Run Triton"))
        self.show_deps_triton_button.setText(_translate("synthesis_view", "Show deps"))
        self.show_ast_triton_button.setText(_translate("synthesis_view", "Show AST"))
        self.run_synthesis_button.setText(_translate("synthesis_view", "Run Synthesis"))
        self.show_ast_synthesis_button.setText(_translate("synthesis_view", "Show AST"))
        self.reassemble_button.setText(_translate("synthesis_view", "Reassemble"))



# built-in module
from __future__ import annotations
from typing import Tuple

# qsynthesis deps
from qsynthesis.tritonast import TritonAst
from qsynthesis.algorithms.synthesizer_base import SynthesizerBase, logger


class TopDownSynthesizer(SynthesizerBase):
    """
    Synthesize with a Top-Down **only** AST search based on Triton AST.
    The complexity in worst case of the search is then O(n) with
    n the number of nodes in the AST to synthesize.
    """

    def synthesize(self, ioast: TritonAst, check_sem: bool = False) -> Tuple[TritonAst, bool]:
        """
        Perform the Top-Down search on the ioast to try synthesizing it. The algorithm first
        tries to synthesize the root node. If not successful descend recursively in all children
        of the AST to try substituting sub-ASTs.

        :param ioast: TritonAst object to synthesize
        :param check_sem: boolean one whether to check the semantic equivalence of expression
                          before substituting them. That ensure soundness of the synthesis
        :returns: tuple with new TritonAst and whether some replacement took place or not

        .. warning:: Activating the `check_sem` parameter implies a strong overhead
                     on the synthesis as SMT queries are being performed for any candidates
        """
        ioast = ioast.duplicate()  # Make a copy of the AST to modify it with breaking the one given in parameter
        self.expr_cache = {}
        self.eval_cache = {}
        self.call_to_eval = 0
        self.eval_count = 0

        next_expr_to_send = None
        has_been_synthesized = False
        is_first = True
        try:
            g = ioast.visit_replacement(update=True)
            while True:
                cur_ast = g.send(next_expr_to_send)
                next_expr_to_send = None  # reset next_expr_to_send

                if not cur_ast.has_children():  # If don't have children either constant or variable
                    continue

                logger.debug(f"try synthesis lookup: {cur_ast.pp_str} [{cur_ast.var_num}]")

                synt_res = self.try_synthesis_lookup(cur_ast, check_sem)

                if synt_res is not None:
                    has_been_synthesized = True
                    logger.debug(f"Replace: {cur_ast.pp_str} ===> {synt_res.pp_str}")
                    if is_first:  # root node has been synthesized
                        return synt_res, has_been_synthesized
                    else:
                        next_expr_to_send = synt_res

                is_first = False

        except StopIteration:
            return ioast, has_been_synthesized



# built-in modules
from enum import Enum
from pathlib import Path
from typing import Tuple, Optional, List, Iterable

# third-party modules
from PyQt5 import QtWidgets, QtCore, QtGui  # provided by IDA

# qsynthesis modules
from qsynthesis.plugin.dependencies import DatabaseManager, Trace, InstrCtx, Instr, Arch, ArchsManager
from qsynthesis.plugin.dependencies import ida_kernwin, IDA_ENABLED, QTRACEIDA_ENABLED, QTRACEDB_ENABLED
from qsynthesis.plugin.dependencies import ida_bytes, ida_nalt, ida_ua, ida_funcs, ida_gdl, ida_loader, ida_lines
from qsynthesis.tables import InputOutputOracleLevelDB, InputOutputOracleREST
from qsynthesis.algorithms import TopDownSynthesizer, PlaceHolderSynthesizer
from qsynthesis.utils.symexec import SimpleSymExec
from qsynthesis.tritonast import ReassemblyError
from qsynthesis.plugin.processor import processor_to_triton_arch, Processor, ProcessorType, processor_to_arch
from qsynthesis.plugin.ast_viewer import AstViewer, BasicBlockViewer
from qsynthesis.plugin.popup_actions import SynthetizeFromHere, SynthetizeToHere, SynthetizeOperand
from qsynthesis.plugin.ui.synthesis_ui import Ui_synthesis_view
from qsynthesis.plugin import slicer
from qsynthesis.types import Addr


TEMPLATE_TRITON = '''<!DOCTYPE html>
<html lang="en">
<head>
    <style>
        body { font-size: large; }
        big { font-size: x-large; }
        td { padding: 5px; padding-left:10px; padding-right:10px; }
    </style>
</head>
<body>
    <center><table>
        <tr><td>Node count</td><td>Depth</td></tr>
        <tr><td align="center"><big><b>%d</b></big></td><td align="center"><big><b>%d</b></big></td></tr>
    </table></center>
    <hr/>
    <center>Inputs
    <table border style="border-style:dashed; margin-top:8px;" >
        %s
    </table>
    </center>
</body>
</html>
'''

TEMPLATE_SYNTHESIS = '''<!DOCTYPE html>
<html lang="en">
<head>
    <style>
        body { font-size: large; }
        big { font-size: x-large; }
        td { padding: 5px; padding-left:10px; padding-right:10px; }
    </style>
</head>
<body>
    <center>
        Simplified: <big style="color:%s;">%s</big><br/>
        Synthesiszed Expression<br/>
        %s
    </center>
    <hr/>
    <center><table>
        <tr><td>Node count</td><td>Depth</td><td>Scale</td></tr>
        <tr>
            <td align="center"><big><b>%d</b></big></td>
            <td align="center"><big><b>%d</b></big></td>
            <td align="center"><big><b>%.2f%c</b></big></td>
        </tr>
    </table></center>
</body>
</html>
'''


class TraceDbType(Enum):
    """ Enum of the Trace type (either database defined in config or a direct sqlite trace """
    CONFIG = 0
    SQLITE = 1


class TargetType(Enum):
    """ Enum of the target expression to be synthesized """
    REG = 0
    MEMORY = 1
    OPERAND = 2


class TableType(Enum):
    """ Enum of the different kind of Lookup Table usable """
    LEVELDB = 0
    HTTP = 1


class AlgorithmType(Enum):
    """ Algorithm type to perform """
    TOPDOWN = "Top-Down"
    PLHDR = "PlaceHolders"


class AnalysisType(Enum):
    """ Type of analysis either based on trace or fully symbolic using Triton only """
    QTRACE = 0
    FULL_SYMBOLIC = 1


class QtraceSymType(Enum):
    """ Mode for trace based DSE. Keeping everything symbolic or just parameters (w.r.t ABI) """
    FULL_SYMBOLIC = 0
    PARAM_SYMBOLIC = 1


class ShowDepState(Enum):
    """ Enum state for buttons showing dependencies """
    SHOW = "Highlight Deps"
    HIDE = "Hide Deps"



# class SynthesizerView(ida_kernwin.PluginForm, Ui_synthesis_view):#QtWidgets.QWidget, ida_kernwin.PluginForm,  Ui_synthesis_view):
class SynthesizerView(ida_kernwin.PluginForm, QtWidgets.QWidget, Ui_synthesis_view):
    """
    Main view of the QSynthesis plugin. Contains all the features implemented
    in the plugin.
    """

    NAME = "QSynthesis"

    def __init__(self, qtrace: Optional['QtraceIDA']):
        """
        Constructor. Take the QtraceIDA object if launched through QtraceIDA
        """
        QtWidgets.QWidget.__init__(self)
        ida_kernwin.PluginForm.__init__(self)
        self.qtrace = qtrace
        self.parent_widget = self

        # Visibility state
        self.closed = True

        # Popup actions
        self.popup_from_here = SynthetizeFromHere(self)
        self.popup_to_here = SynthetizeToHere(self)
        self.popup_operand = SynthetizeOperand(self)

        # Analysis variables
        self.symexec = None
        self.lookuptable = None
        self.ast = None
        self.synthesizer = None
        self.synth_ast = None
        # For operand synthesis
        self.stop_addr = None  # Address where to analysis end
        self.op_num = None
        self.op_is_read = False  # Negation of if the operand is written

        # If working on its own
        self._dbm = None
        self._trace = None
        self._arch = processor_to_arch()  # By default initialize it with current architecture

        # Expresssion highlighted
        self.highlighted_addr = {}  # addr -> backed_color

    @property
    def arch(self) -> Arch:
        """ Return current QtraceDB Arch object whether it comes from qtraceida or the trace directly """
        if QTRACEIDA_ENABLED and self.qtrace_mode:
            return self.qtrace.arch
        else:
            return self._arch

    @property
    def trace(self) -> Trace:
        """ Return current Trace object whether it comes from qtraceida or a manual loading """
        if QTRACEIDA_ENABLED and self.qtrace_mode:
            return self.qtrace.trace
        else:
            return self._trace

    @property
    def qtrace_mode(self) -> bool:
        """ Return whether QSynthesis was launched through qtrace or not """
        return self.qtrace is not None

    def get_instruction(self, ea: Addr) -> Instr:
        """
        Disassemble the current instruction as a QtraceDB Instruction object

        :param ea: address of the head of the instruction
        :return: instruction object

        .. warning:: The current address have to be code and being
                     the head of an instruction
        """
        opc = ida_bytes.get_bytes(ea, ida_bytes.get_item_size(ea))
        return self.arch.disasm_one(opc, ea)

    def is_lookuptable_loaded(self) -> bool:
        """ Return true if the lookup table is loaded """
        return self.lookuptable is not None

    def OnCreate(self, form) -> None:
        """
        PluginForm callback called when view is created. It initialize
        the whole view with all the widgets.
        """
        self.parent_widget = self.FormToPyQtWidget(form)
        # Init of the view has to be done here ?
        self.init()

        # Put that check here otherwise init will override this enabling
        # NOTE: Might be better to implement it in a proprer callback
        if self.trace:
            self.on_trace_opened(self.trace)
        else:
            self.set_algorithm_type_enable(AnalysisType.QTRACE, False)

    def Show(self) -> bool:
        """ Creates the form if not created or focuses it if it was """
        self.closed = False
        self.enable_popups()
        opts = ida_kernwin.PluginForm.WOPN_PERSIST
        r = ida_kernwin.PluginForm.Show(self, self.NAME, options=opts)
        ida_kernwin.set_dock_pos(self.NAME, "IDA View-A", ida_kernwin.DP_RIGHT)
        return r

    def OnClose(self, form) -> None:
        """ Window closed callback. Disable all popup actions """
        # Change visibility state
        print("qsynthesis closed")
        self.closed = True
        self.disable_popups()

    def enable_popups(self) -> None:
        """ Enable IDA View popup actions """
        self.popup_from_here.register()
        self.popup_to_here.register()
        self.popup_operand.register()

    def disable_popups(self) -> None:
        """ Disabled IDA View popup actions """
        self.popup_from_here.unregister()
        self.popup_to_here.unregister()
        self.popup_operand.unregister()

    def set_visible_all_layout(self, layout: QtWidgets.QLayout, val: bool) -> None:
        """
        Change the visibility of all the widgets contained in a layout

        :param layout: Layout widget in which to show/hide all sub-components
        :param val: Boolean value to either set visible or invisible
        :return: None
        """
        for i in range(layout.count()):
            item = layout.itemAt(i)
            item_w = item.widget()
            if item_w is None:
                if isinstance(item, QtWidgets.QLayout):
                    self.set_visible_all_layout(item, val)
            else:
                item_w.setVisible(val)

    def set_enabled_all_layout(self, layout: QtWidgets.QLayout, val: bool) -> None:
        """
        Change the 'enabled' state of all the widgets contained in a layout

        :param layout: Layout widget in which to enable/disable all sub-components
        :param val: Boolean value to either set enabled or disabled
        :return: None
        """
        for i in range(layout.count()):
            item = layout.itemAt(i)
            item_w = item.widget()
            if item_w is None:
                if isinstance(item, QtWidgets.QLayout):
                    self.set_enabled_all_layout(item, val)
            else:
                item_w.setEnabled(val)

    def init(self) -> None:
        """
        Initialization function called on view creation to instanciate
        all widgets and its default state.
        """
        self.setupUi(self.parent_widget)
        if QTRACEIDA_ENABLED or not QTRACEDB_ENABLED:  # The trace will be provided by qtraceida so not showing this
            self.set_visible_all_layout(self.traceLayout, False)
        else:  # Initialize the fields and actions
            self.trace_type_box.addItems([x.name for x in TraceDbType])
            self.trace_type_box.currentIndexChanged.connect(self.trace_type_changed)
            self.trace_line.focusInEvent = self.customfocusInEventTraceLine

        # Target configuration
        reg = QtCore.QRegExp("^(0x)?[a-fA-F0-9]+$")
        validator = QtGui.QRegExpValidator(reg)
        self.from_line.setValidator(validator)
        self.from_line.textChanged.connect(self.from_line_changed)
        self.from_line.editingFinished.connect(self.from_line_reset_style)
        self.to_line.textChanged.connect(self.to_line_changed)
        self.to_line.editingFinished.connect(self.to_line_reset_style)
        self.to_line.setValidator(validator)
        self.mem_line.textChanged.connect(self.mem_line_changed)
        self.mem_line.editingFinished.connect(self.mem_line_reset_style)
        self.mem_line.setValidator(validator)
        self.target_box.addItems([x.name for x in TargetType])
        self.target_box.currentIndexChanged.connect(self.target_changed)
        self.mem_line.setVisible(False)
        self.operand_label.setVisible(False)

        # Table configuration
        self.table_type_box.addItems([x.name for x in TableType])
        self.table_type_box.currentIndexChanged.connect(self.table_type_changed)
        self.table_line.focusInEvent = self.customfocusInEventLookupTableLine

        # Algorithm configuration
        self.algorithm_box.addItems([x.value for x in AlgorithmType])

        # Add both QTRACE, and FULL_SYMBOLIC algorithms and disabled them
        self.algorithm_type_box.addItems([x.name for x in AnalysisType])
        self.algorithm_type_box.currentIndexChanged.connect(self.algorithm_type_changed)
        self.set_algorithm_type_enable(AnalysisType.QTRACE, QTRACEDB_ENABLED)
        self.set_algorithm_type_enable(AnalysisType.FULL_SYMBOLIC, IDA_ENABLED)
        self.qtrace_sym_type_box.addItems([x.name for x in QtraceSymType])

        # Hide all experimental settings
        self.set_visible_all_layout(self.experimentalLayout, False)

        self.run_triton_button.clicked.connect(self.run_triton_clicked)
        self.show_deps_triton_button.setText(ShowDepState.SHOW.value)
        self.show_deps_triton_button.clicked.connect(self.triton_show_deps_clicked)
        self.show_ast_triton_button.clicked.connect(self.triton_show_ast_clicked)
        self.run_synthesis_button.clicked.connect(self.run_synthesis_clicked)
        self.show_ast_synthesis_button.clicked.connect(self.synthesis_show_ast_clicked)
        self.reassemble_button.clicked.connect(self.reassemble_clicked)

    def switch_to_target_operand(self) -> None:
        """ Change target type to operand. (Meant to be called by the popup action) """
        self.target_box.setCurrentIndex(TargetType.OPERAND.value)

    def set_algorithm_type_enable(self, alg: AnalysisType, enabled: bool):
        # We assume here that items are in order (with their value attribute)
        self.algorithm_type_box.model().item(alg.value).setEnabled(enabled)
        if not enabled:
            self.algorithm_type_box.setCurrentIndex(AnalysisType.FULL_SYMBOLIC.value if AnalysisType.QTRACE else AnalysisType.QTRACE.value)
            self.algorithm_type_changed()

    def algorithm_type_changed(self) -> None:
        """
        When changing algorithm type. Enable or disable the mode
        associated with Qtrace.
        """
        self.qtrace_sym_type_box.setEnabled(self.analysis_type == AnalysisType.QTRACE)

    @property
    def qtrace_sym_type(self) -> QtraceSymType:
        """ Return the current symbolic mode of analysis for trace DSE """
        return QtraceSymType(self.qtrace_sym_type_box.currentIndex())

    def from_line_reset_style(self) -> None:
        """ Reset from_line stylesheet """
        self.from_line.setStyleSheet("")

    def from_line_changed(self) -> None:
        """
        Callback called when something change in 'From' field. Validated it
        against the validator (and either show red / green around the box)
        """
        color = "green" if self.from_line.hasAcceptableInput() else "red"
        self.from_line.setStyleSheet(f"border: 1px solid {color}")

    def to_line_reset_style(self) -> None:
        """ Reset to_line stylesheet """
        self.to_line.setStyleSheet("")

    def to_line_changed(self) -> None:
        """
        Callback called when something change in 'To' field. Validated it
        against the validator (and either show red / green around the box)
        """
        color = "green" if self.to_line.hasAcceptableInput() else "red"
        self.to_line.setStyleSheet(f"border: 1px solid {color}")

    def mem_line_reset_style(self) -> None:
        """ Reset the memory line stylesheet """
        self.mem_line.setStyleSheet("")

    def mem_line_changed(self) -> None:
        """
        Callback called when something change in 'Memory' field. Validated it
        against the validator (and either show red / green around the box)
        """
        color = "green" if self.mem_line.hasAcceptableInput() else "red"
        self.mem_line.setStyleSheet(f"border: 1px solid {color}")

    def open_no_trace(self) -> None:
        """
        Function called when the plugin is opened wihtout any active
        trace.
        """
        self.on_trace_opened(None)
        self.set_algorithm_type_enable(AnalysisType.QTRACE, False)

    def on_trace_opened(self, trace: Optional[Trace]) -> None:
        """
        Callback called when a trace is opened either via Qtrace-IDA
        or directly when Qtrace-IDA is not available.

        :param trace: Trace just having been opened
        """
        if trace:
            print(f"QSynthesis: on_trace_opened ({trace})")
        self.set_algorithm_type_enable(AnalysisType.QTRACE, True)
        # Activate all configuration lines
        self.set_enabled_all_layout(self.targetLayout, True)
        # Initialize registers of the targetcombobox
        self.register_box.addItems([x.name for x in ArchsManager.get_supported_regs(self.arch)])
        self.set_enabled_all_layout(self.tableLayout, True)
        self.set_enabled_all_layout(self.algorithmLayout, True)
        self.run_triton_button.setEnabled(True)

    def target_changed(self, index: int) -> None:
        """ Enable and disable appropriate widget when changing the target selector """
        t = TargetType(index)
        if t == TargetType.REG:
            self.register_box.setVisible(True)
            self.mem_line.setVisible(False)
            self.operand_label.setVisible(False)
        elif t == TargetType.MEMORY:
            self.register_box.setVisible(False)
            self.operand_label.setVisible(False)
            self.mem_line.setVisible(True)
        elif t == TargetType.OPERAND:
            self.register_box.setVisible(False)
            self.mem_line.setVisible(False)
            self.operand_label.setVisible(True)
            self.operand_label.clear()
        else:
            assert False

    @property
    def target_type(self) -> TargetType:
        """ Return target type as an enum """
        return TargetType(self.target_box.currentIndex())

    @property
    def table_type(self) -> TableType:
        """ Return table type as an enum """
        return TableType(self.table_type_box.currentIndex())

    def table_type_changed(self, _) -> None:
        """ Called when table type changed. (just set a placeholder string) """
        self.table_line.setText("")
        if self.table_type == TableType.HTTP:
            self.table_line.setPlaceholderText('e.g: http://localhost')
        else:
            self.table_line.setPlaceholderText("")

    def customfocusInEventLookupTableLine(self, event) -> None:
        """
        Callback called when to focus is given to the table field.
        Instead of leaving the user writing a path, open a dialog
        to open a given file.
        """
        self.table_line.setStyleSheet("")
        if self.table_type == TableType.LEVELDB:
            filename = QtWidgets.QFileDialog.getExistingDirectory()
            filepath = Path(filename)
            if filepath.exists() and filepath.is_dir():
                self.table_line.setText(str(filepath))
            else:
                print(f"Invalid file: {filepath}")

            self.trace_line.focusNextChild()

    @property
    def algorithm(self) -> AlgorithmType:
        """ Return algorithm type as an enum """
        return AlgorithmType(self.algorithm_box.currentText())

    @property
    def analysis_type(self) -> AnalysisType:
        """ Return analysis type as an enum """
        return AnalysisType[self.algorithm_type_box.currentText()]

    def run_triton_clicked(self) -> None:
        """
        Triton button clicked callback. Verifies that all parameters are
        valid before triggering the symbolic execution. At the end delegate
        the computation to :meth:`SynthesizerView.run_triton_qtrace` or
        :meth:`SynthesizerView.run_triton_fullsym`.
        """
        self.ast = None  # Reset the AST variable
        self.triton_textarea.clear()  # clear
        self.set_enabled_synthesis_widgets(False)

        if self.is_lookuptable_loaded():
            if self.lookuptable.name != self.table_line.text():
                print("Reload lookup table")
                ret = self.load_lookup_table()
            else:
                ret = True
        else:
            ret = self.load_lookup_table()
        if not ret:
            return

        if self.analysis_type == AnalysisType.QTRACE:
            ret = self.run_triton_qtrace()
        elif self.analysis_type == AnalysisType.FULL_SYMBOLIC:
            ret = self.run_triton_fullsym()
        else:
            assert False

        if ret:
            self.on_triton_finished()

    def load_lookup_table(self) -> bool:
        """
        Open the lookup table selected. If LevelDB open database and if REST
        checks that the remote URL is reachable.

        :return: True if the opening succeeded
        """
        if not self.table_line.text():
            self.table_line.setStyleSheet("border: 1px solid red")
            return False
        if self.table_type == TableType.HTTP:
            try:
                self.lookuptable = InputOutputOracleREST.load(self.table_line.text())
            except ConnectionAbortedError as e:
                QtWidgets.QMessageBox.critical(self, "Table Loading", f"Error contacting {self.table_line}\n{e}")
                return False
        elif self.table_type == TableType.LEVELDB:
            try:
                self.lookuptable = InputOutputOracleLevelDB.load(self.table_line.text())
            except Exception as e:
                QtWidgets.QMessageBox.critical(self, "Table Loading", f"Error when loading database: {e}")
                return False
        return True

    def line_to_qtrace_inst(self, widget: QtWidgets.QLineEdit) -> Optional[InstrCtx]:
        """
        Retrieve a trace instruction from a LineEdit widget which content
        is meant to be an address.

        :param widget: LineEdit widget containing an address (in our context from_line and to_line)
        :return: First instruction in the trace hitting the address (if valid)
        """
        if not widget.hasAcceptableInput():
            widget.setStyleSheet("border: 1px solid red")
            return None
        else:
            addr = int(widget.text(), 16)
            inst = self.trace.get_first_instr_at_addr(addr)
            if inst is None:
                QtWidgets.QMessageBox.critical(self, "Invalid parameter", f"No instruction in trace at address: {addr:#x}")
                return None
            else:
                return inst

    def run_triton_qtrace(self) -> bool:
        """
        Run the symbolic execution with Triton on the Trace. First performs
        all the sanitization on the fields and then delegate computation to
        the `QtraceSymExec` object of qsynthesis.utils.

        :return: True if the symbolic execution procceeded without errors
        """
        # Make sure a trace is loaded before proceeding
        if self.trace is None:
            QtWidgets.QMessageBox.critical(self, "No trace", f"A trace should be loaded first ")
            return False

        from_inst = self.line_to_qtrace_inst(self.from_line)
        if from_inst is None:
            return False
        to_inst = self.line_to_qtrace_inst(self.to_line)
        if to_inst is None:
            return False
        if to_inst.id < from_inst.id:
            QtWidgets.QMessageBox.critical(self, "Invalid order", f"Instruction {from_inst.addr:#x}{from_inst} higher in the trace than {to_inst.addr:#x}{to_inst} ({from_inst.id} > {to_inst.id})")
            return False

        if self.target_type == TargetType.MEMORY:
            if not self.mem_line.hasAcceptableInput():
                self.mem_line.setStyleSheet("border: 1px solid red")
                return False
            else:
                mem_addr = int(self.mem_line.text(), 16)

        # Do not execute last instruction if operand & read operand else always execute last instruction
        off = 0 if self.target_type == TargetType.OPERAND and self.op_is_read else 1

        # Set the stop addr (for operand reassembly)
        self.stop_addr = to_inst.addr

        # At this point we are sur to have valid parameters

        from qsynthesis.utils.qtrace_symexec import QtraceSymExec, Mode
        m = Mode[self.qtrace_sym_type.name]
        self.symexec = QtraceSymExec(self.trace, m)
        self.symexec.initialize_register(self.arch.INS_PTR.name, from_inst.INS_PTR)
        self.symexec.initialize_register(self.arch.STK_PTR.name, from_inst.STK_PTR)
        self.symexec.process_instr_sequence(from_inst.id, to_inst.id+off)
        if self.target_type == TargetType.REG:
            self.ast = self.symexec.get_register_ast(self.register_box.currentText())
        elif self.target_type == TargetType.MEMORY:
            self.ast = self.symexec.get_memory_ast(mem_addr, int(self.lookuptable.bitsize/8))
        elif self.target_type == TargetType.OPERAND:
            if self.op_is_read: # if read disassemble only the last instruction
                inst = self.symexec.disassemble(to_inst.opcode, to_inst.addr)
                self.ast = self.symexec.get_operand_ast(self.op_num, inst)
            else:  # operand is write can directly get its ast
                self.ast = self.symexec.get_operand_ast(self.op_num)
        else:
            assert False
        self.symexec.ctx.clearCallbacks()  # Fix the bug from space / can also be fixed by making self.symexec object attribute
        return True

    def run_triton_fullsym(self) -> bool:
        """
        Run the symbolic execution in 'pure' symbolic (without trace). First performs
        all the sanitization on the fields and then delegate computation to the
        `SimpleSymExec` object of qsynthesis.utils.

        :return: True if the symbolic execution procceeded without errors
        """
        if Processor.type == ProcessorType.UNKNOWN:
            QtWidgets.QMessageBox.critical(self, "Unsupported arch", f"Architecture {Processor.name} is not supported by Triton")
            return False
        if not self.from_line.hasAcceptableInput():
            self.from_line.setStyleSheet("border: 1px solid red")
            return False
        if not self.to_line.hasAcceptableInput():
            self.to_line.setStyleSheet("border: 1px solid red")
            return False
        cur_addr = int(self.from_line.text(), 16)
        end_addr = int(self.to_line.text(), 16)
        if end_addr <= cur_addr:
            QtWidgets.QMessageBox.critical(self, "Invalid order", f"From {cur_addr:#x} must be lower than {end_addr:#x}")
            return False
        if not ida_bytes.is_mapped(cur_addr) or not ida_bytes.is_mapped(end_addr):
            QtWidgets.QMessageBox.critical(self, "Invalid address",
                                           f"From: {cur_addr:#x} or To: {end_addr:#x} is not mapped in memory")
            return False

        if self.target_type == TargetType.MEMORY:
            if not self.mem_line.hasAcceptableInput():
                self.mem_line.setStyleSheet("border: 1px solid red")
                return False
            else:
                mem_addr = int(self.mem_line.text(), 16)

        if self.target_type == TargetType.OPERAND and self.op_is_read:
            self.stop_addr = end_addr  # Do not execute last address
        else:
            self.stop_addr = ida_bytes.get_item_end(end_addr)  # Do execute 'to' address
        # At this point we are sur to have valid parameters

        # Create the purely symbolic executor
        self.symexec = SimpleSymExec(processor_to_triton_arch())
        from triton import MODE
        self.symexec.ctx.setMode(MODE.ONLY_ON_SYMBOLIZED, False)

        self.symexec.initialize_register(self.arch.INS_PTR.name, cur_addr)
        self.symexec.initialize_register(self.arch.STK_PTR.name, 0x800000)

        # Execute the range of instructions
        while cur_addr < self.stop_addr:  # Retrieve directly bytes from IDA
            if ida_bytes.is_code(ida_bytes.get_flags(cur_addr)):
                opc = ida_bytes.get_bytes(cur_addr, ida_bytes.get_item_size(cur_addr))
                if not self.symexec.execute(opc):
                    QtWidgets.QMessageBox.critical(self, "Symbolic Execution Error", f"Instruction at address {cur_addr:#x} seems unsupported by Triton")
                    return False
            else:
                QtWidgets.QMessageBox.critical(self, "Invalid byte", f"Stop on address: {cur_addr:#x} which is not code")
                return False
            cur_addr = ida_bytes.next_head(cur_addr, self.stop_addr)

        if self.target_type == TargetType.REG:
            self.ast = self.symexec.get_register_ast(self.register_box.currentText())
        elif self.target_type == TargetType.MEMORY:
            self.ast = self.symexec.get_memory_ast(mem_addr, int(self.lookuptable.bitsize/8))
        elif self.target_type == TargetType.OPERAND:
            if self.op_is_read: # if read disassemble only the last instruction
                opc = ida_bytes.get_bytes(self.stop_addr, ida_bytes.get_item_size(self.stop_addr))
                inst = self.symexec.disassemble(opc, self.stop_addr)
                self.ast = self.symexec.get_operand_ast(self.op_num, inst)
            else:  # operand is write can directly get its ast
                self.ast = self.symexec.get_operand_ast(self.op_num)
        else:
            assert False

        self.symexec.ctx.clearCallbacks()
        return True

    def on_triton_finished(self) -> None:
        """
        Function called when symbolic execution terminated successfully.
        Enable all the synthesis related widget in the view.
        """
        # Enable Triton fields
        self.triton_textarea.setEnabled(True)
        if IDA_ENABLED:
            self.show_deps_triton_button.setEnabled(True)
        if QTRACEIDA_ENABLED:
            self.show_ast_triton_button.setEnabled(True)

        # Fill the text area with the results
        self.fill_triton_results()

        # Enable all buttons related to synthesis
        self.set_enabled_synthesis_widgets(True)

    def fill_triton_results(self) -> None:
        """ Generate the resulting analysis HTML and fill the widget """
        if self.ast.symvars:
            tpl = '<tr><td align="center">%s</td><td align="center">%d</td></tr>'
            inp_s = "\n".join(tpl % (x.getAlias(), x.getBitSize()) for x in self.ast.symvars)
        else:
            inp_s = "<big><b>0</b></big>"
        self.triton_textarea.setHtml(TEMPLATE_TRITON % (self.ast.node_count, self.ast.depth, inp_s))

    def triton_show_deps_clicked(self) -> None:
        """
        Callback called when Show Dependencies is clicked. The function
        retrieve the dependencies (pre-computed during symbolic execution)
        and color the appropriate lines in the IDA view.
        """
        st = ShowDepState(self.show_deps_triton_button.text())
        if st == ShowDepState.SHOW:
            addrs = self.get_dependency_addresses()
            for addr in addrs:
                back = ida_nalt.get_item_color(addr)
                self.highlighted_addr[addr] = back
                ida_nalt.set_item_color(addr, 0xA1F7A1)

        else:
            for addr, color in self.highlighted_addr.items():
                ida_nalt.set_item_color(addr, color)
            self.highlighted_addr.clear()

        # Switch state
        self.show_deps_triton_button.setText(ShowDepState.HIDE.value if st == ShowDepState.SHOW else ShowDepState.SHOW.value)

    def get_dependency_addresses(self) -> List[Addr]:
        """
        Retrieve Triton 'SymbolicExpression' of the expression to synthesize as
        'SymbolicExpression' contain a comment field which contain the link between
        expressions and Instructions. The expression is given to the slicer iterate
        the expression to retrieve the set of instruction involved in computation.

        :return: list of addresses involved in the computation of the expression
        """
        if self.target_type == TargetType.REG:
            sym = self.symexec.get_register_symbolic_expression(self.register_box.currentText())
        elif self.target_type == TargetType.MEMORY:
            # FIXME: Sanitize memory value before using it
            mem_addr = int(self.mem_line.text(), 16)
            sym = self.symexec.get_memory_symbolic_expression(mem_addr, int(self.lookuptable.bitsize / 8))
        elif self.target_type == TargetType.OPERAND:
            sym = self.symexec.get_operand_symbolic_expression(self.op_num)
        else:
            assert False

        # Iterate all addresses
        return sorted(slicer.backslice(sym))

    def triton_show_ast_clicked(self) -> None:
        """ Show the AST of the triton expression """
        viewer = AstViewer("Triton AST", self.ast)
        viewer.Show()

    def run_synthesis_clicked(self) -> None:
        """
        Run the synthesis using the algorithm selected.
        """

        # Check the compatibiliy
        if not self.lookuptable.is_expr_compatible(self.ast):
            QtWidgets.QMessageBox.critical(self, "Incompatibility",
                                           f"Variables of the AST expression {self.ast.symvars} are incompatibles"
                                           f"with table variables {self.lookuptable.grammar.vars_dict} (in sizes)")
            return

        if self.algorithm == AlgorithmType.TOPDOWN:
            synthesizer = TopDownSynthesizer(self.lookuptable)
        else:
            synthesizer = PlaceHolderSynthesizer(self.lookuptable)
        self.synth_ast, _ = synthesizer.synthesize(self.ast)

        self.on_synthesis_finished()

    def on_synthesis_finished(self) -> None:
        """
        Upon synthesis termination. Enable the remaining buttons
        """
        # Enable Synthesis fields
        self.synthesis_textarea.setEnabled(True)
        if QTRACEIDA_ENABLED:
            self.show_ast_synthesis_button.setEnabled(True)
        if IDA_ENABLED:
            self.reassemble_button.setEnabled(True)

        # Fill the text area with the results
        self.fill_synthesis_results()

    def fill_synthesis_results(self) -> None:
        """
        Fill synthesis result widget as pre-formatted HTML
        """
        color, simp = ("green", "Yes") if self.ast.node_count > self.synth_ast.node_count else ("red", "No")
        ssexpr = "<small>(too large)</small>" if self.synth_ast.node_count > 12 else f"<big><b>{self.synth_ast.pp_str}</b></big>"
        scale = -(((self.ast.node_count - self.synth_ast.node_count) * 100) / self.ast.node_count)
        self.synthesis_textarea.setHtml(TEMPLATE_SYNTHESIS % (color, simp, ssexpr, self.synth_ast.node_count, self.synth_ast.depth, scale, '%'))

    def synthesis_show_ast_clicked(self) -> None:
        """ Show the AST of the synthesized expression """
        viewer = AstViewer("Synthesized AST", self.synth_ast)
        viewer.Show()

    def set_enabled_synthesis_widgets(self, val: bool) -> None:
        """
        Enable or Disable synthesis related widgets

        :param val: boolean on whether to enable or disable the widgets
        """
        self.run_synthesis_button.setEnabled(val)
        self.synthesis_textarea.setEnabled(val)
        if not val:  # In case we disable everything
            self.synthesis_textarea.clear()  # Clear synthesis view
            self.show_ast_synthesis_button.setEnabled(val)
            self.reassemble_button.setEnabled(val)

    def reassemble_clicked(self) -> None:
        """
        Callback clicked when Reassembly is clicked. Shows an additional
        window to select various parameters and the call the appropriate
        auxiliary function.
        """
        is_reg, reg = self.selected_expr_register()
        res = self.get_reassembly_options(ask_reg=not is_reg)
        if res:
            rg, patch_fun, shrink_fun, snap = res
            dst_reg = reg.lower() if is_reg else rg.lower()
            try:
                if patch_fun:
                    addrs = self.get_dependency_addresses()
                    asm_bytes = self.synth_ast.reassemble(dst_reg, self.arch.NAME)
                    if snap:  # Create a snapshot of the database
                        ss = ida_loader.snapshot_t()
                        ss.desc = f"Reassembly of {dst_reg} at {self.stop_addr:#x}"
                        ida_kernwin.take_database_snapshot(ss)
                    if shrink_fun:
                        self.patch_and_shrink_reassembly(addrs, asm_bytes)
                    else:
                        self.patch_reassembly(addrs, asm_bytes)
                else:
                    # Reassemble to instruction and show it in a View
                    raw_insts = self.synth_ast.reassemble(dst_reg, self.arch.NAME)
                    insts = self.arch.disasm(raw_insts, 0x0)
                    bb_viewer = BasicBlockViewer("Reassembly", insts)
                    bb_viewer.Show()
            except ReassemblyError as e:
                QtWidgets.QMessageBox.critical(self, "Reassembly error", f"Error: {e}")
        else:
            pass # Do nothing user click canceled on options

    @staticmethod
    def coallesce_addrs(addrs: Iterable[Addr]) -> List[Tuple[Addr, int]]:
        """ Convert list of addresses into blocks of addr+size """
        blocks = []
        for addr in addrs:
            sz = ida_bytes.get_item_size(addr)
            if not blocks:
                blocks.append((addr, sz))
            else:
                b_addr, b_sz = blocks[-1]
                if b_addr + b_sz == addr:  # We are next to the previous item
                    blocks[-1] = (b_addr, b_sz+sz)
                else:  # The address is not contiguous
                    blocks.append((addr, sz))
        return blocks

    def patch_reassembly(self, addrs: List[Addr], asm: bytes) -> None:
        """
        Patch given addresses with NOPs and put `asm` bytes corresponding
        to reassembled instruction on the last addresses to put reassembled
        byte near the location the expression was extracted from.

        :param addrs: sorted list of addresses to strip
        :param asm: reassembled instructino bytes
        """
        nop = self.arch.nop_instruction
        insts = self.arch.disasm(asm, 0x0)
        blocks = self.coallesce_addrs(addrs)
        if sum(x[1] for x in blocks) < len(asm):
            QtWidgets.QMessageBox.critical(self, "Reassembly error", f"No enough place to push back reassembled instructions")
            return

        block_addr, block_sz = blocks.pop()
        for i in insts[::-1]:
            data = i.bytes
            while len(data) > block_sz:  # While blocks are too small fill them with nop
                if block_sz != 0:
                    ida_bytes.patch_bytes(block_addr, nop * block_sz)  # Fill space of the block with NOPs
                if blocks:
                    block_addr, block_sz = blocks.pop()
                else:
                    QtWidgets.QMessageBox.critical(self, "Reassembly error", "No slot to place instruction remaining, abort")
                    return

            # Here there is meant to be enough space to place instruction
            p_addr = block_addr + block_sz - len(data)  # compute address where to patch within block
            self.safe_patch_instruction(p_addr, data)
            block_sz -= len(data)

        # All instructions have been put in blocks, fill the remaining ones with NOPs
        while blocks or block_sz:
            if block_sz:
                ida_bytes.patch_bytes(block_addr, nop * block_sz)  # Fill space of the block with NOPs
                block_sz = 0
            if blocks:
                block_addr, block_sz = blocks.pop()

    @staticmethod
    def safe_patch_instruction(ea: Addr, data: bytes) -> None:
        """
        Patch a single instruction with its bytes (data). Make sure IDA
        recognize it as an instruction after patching.

        :param ea: address to patch
        :param data: instruction bytes
        """
        ida_bytes.patch_bytes(ea, data)  # Patch bytes
        if not ida_bytes.is_code(ida_bytes.get_flags(ea)):  # Check that the address is now a code instruction, if not:
            ida_bytes.del_items(ea, 0, len(data))   # Del all types
            ida_ua.create_insn(ea)                  # Redecode instruction
        if not ida_bytes.is_code(ida_bytes.get_flags(ea)):
            print("Really can't create instruction at that location")

    @staticmethod
    def safe_patch_instruction_block(ea: Addr, data: bytes) -> None:
        """
        Patch an address with a bunch of instruction bytes (data). Make sure
        IDA recognize all of them as instruction after patching.

        :param ea: address where to patch
        :param data: instruction bytes to put as patch
        """
        print(f"safe_patch_instruction_block {ea:#x}: {data}")
        ida_bytes.patch_bytes(ea, data)  # Patch bytes
        ida_bytes.del_items(ea, 0, len(data))
        ida_ua.create_insn(ea)

    def patch_and_shrink_reassembly(self, addrs, asm) -> None:
        init_addr = addrs[0]
        f = ida_funcs.get_func(init_addr)
        f_start_ea = f.start_ea
        g = ida_gdl.FlowChart(f)
        low, high = init_addr, addrs[-1]
        block = None
        for bb in g:
            if bb.start_ea <= low < bb.end_ea and bb.start_ea <= high <= bb.end_ea:
                block = bb
        if block is None:
            QtWidgets.QMessageBox.critical(self, "Reassembly error", "Dependency slice have to be in the same basic block")
            return

        # Scan all items
        cur_addr = init_addr
        payload = b""
        while cur_addr < block.end_ea:  # Iterate the whole basicblock
            if not addrs and asm: # if we poped all the dependencies we can put the synthesized expr
                payload += asm
                asm = None
            
            if cur_addr not in addrs:  # Instruction not in dependency so keep it
                sz = ida_bytes.get_item_size(cur_addr)
                payload += ida_bytes.get_bytes(cur_addr, sz)
            else:
                addrs.remove(cur_addr)  # remove the address from deps

            cur_addr = ida_bytes.next_head(cur_addr, block.end_ea)

        if asm:  # If it has not been "Noned" it has not been put in payload
            print("reassembled payload has not been placed in the final basic bloc (abort reassembly)")
            return

        # Perform the final patching
        self.safe_patch_instruction_block(init_addr, payload)

        # If the block was the first of the function and we rewrote the head
        # and the function has been destroyed. Reconstruct it.
        if init_addr == f_start_ea and ida_funcs.get_func(init_addr) is None:
            res = ida_funcs.add_func(init_addr)
            print(f"Function reconstruction: {res}")
            return

        # If the block was the last of the function (adjust the end of the function)
        if f.end_ea == block.end_ea:
            print(f"adjust of the function to {init_addr+len(payload):#x}")
            ida_funcs.set_func_end(init_addr, init_addr+len(payload))

    def get_reassembly_options(self, ask_reg: bool = False) -> Optional[Tuple[str, bool, bool, bool]]:
        """
        Function showing a Dialog to ask reassembly parameters.

        :param ask_reg: boolean on whether to ask the destination register or not
        :return: optional list of options
        """
        dlg = QtWidgets.QDialog(parent=None)
        dlg.setWindowTitle('Reassembly options')

        v_layout = QtWidgets.QVBoxLayout(dlg)
        if ask_reg:
            h_layout = QtWidgets.QHBoxLayout()
            label_reg = QtWidgets.QLabel(dlg)
            label_reg.setText("Destination register:")
            h_layout.addWidget(label_reg)
            combobox = QtWidgets.QComboBox(dlg)
            combobox.addItems([x.name for x in ArchsManager.get_supported_regs(self.arch)])
            h_layout.addWidget(combobox)
            v_layout.addLayout(h_layout)

        patch_fun = QtWidgets.QCheckBox(dlg)
        patch_fun.setText("patch function bytes")
        v_layout.addWidget(patch_fun)

        shrink_fun = QtWidgets.QCheckBox(dlg)
        shrink_fun.setText("shrink function\n move some instruction instead of filling with NOPs.\nCan break disassembly"
                           " for relative instructions. (Works only for linear blocks)")
        shrink_fun.setEnabled(False)
        v_layout.addWidget(shrink_fun)

        def patch_checked():
            if patch_fun.isChecked():
                shrink_fun.setEnabled(True)
            else:
                shrink_fun.setEnabled(False)
                shrink_fun.setChecked(False)

        patch_fun.stateChanged.connect(patch_checked)

        snapshot = QtWidgets.QCheckBox(dlg)
        snapshot.setText("Snapshot database before patching")
        v_layout.addWidget(snapshot)

        buttonbox = QtWidgets.QDialogButtonBox(dlg)
        buttonbox.setOrientation(QtCore.Qt.Horizontal)
        buttonbox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel | QtWidgets.QDialogButtonBox.Ok)
        v_layout.addWidget(buttonbox)
        buttonbox.accepted.connect(dlg.accept)
        buttonbox.rejected.connect(dlg.reject)

        dlg.exec()

        if dlg.result():
            reg = combobox.currentText() if ask_reg else None
            return reg, patch_fun.isChecked(), shrink_fun.isChecked(), snapshot.isChecked()
        else:
            return None

    def selected_expr_register(self) -> Tuple[bool, Optional[str]]:
        """
        Return the currently selected register depending on target type.
        None if no register is selected
        """
        if self.target_type == TargetType.REG:
            return True, self.register_box.currentText()
        elif self.target_type == TargetType.MEMORY:
            return False, None
        elif self.target_type == TargetType.OPERAND:

            opc = ida_bytes.get_bytes(self.stop_addr, ida_bytes.get_item_size(self.stop_addr))
            inst = self.arch.disasm_one(opc, self.stop_addr)
            if inst is None:
                return False, None
            op = inst.operands[self.op_num]
            if op.is_register():
                op_name = ida_lines.tag_remove(ida_ua.print_operand(self.stop_addr, self.op_num))
                return True, op_name  # op.register.name
            else:
                return False, None

    # =====================  Utility functions when Qtrace-IDA is NOT available  ======================
    @property
    def trace_type(self) -> TraceDbType:
        """ Trace type to load as an enum """
        return TraceDbType(self.trace_type_box.currentIndex())

    def trace_type_changed(self, _):
        """ Trace field changed """
        self.trace_line.setText("")

    def customfocusInEventTraceLine(self, event):
        """ Triggered when getting the focus on trace line edit. To help to user selecting a trace. """
        self.trace_line.setText("")
        if self.trace_type == TraceDbType.CONFIG:
            self._dbm = DatabaseManager.from_qtracedb_config()
            t_name = self.get_trace_from_db()
            if t_name:
                self._trace = self._dbm.get_trace(t_name)
                self._arch = self._trace.get_arch()
                self.trace_line.setText(t_name)
                self.on_trace_opened(self._trace)
        elif self.trace_type == TraceDbType.SQLITE:
            filename = QtWidgets.QFileDialog.getOpenFileName()[0]
            filepath = Path(filename)
            if filepath.exists() and filepath.is_file():
                try:
                    self._dbm = DatabaseManager(f'sqlite:///{filepath.absolute()}')
                    self._trace = self._dbm.get_trace("x86_64")
                    self._arch = self._trace.get_arch()
                    self.trace_line.setText(str(filepath))
                    self.on_trace_opened(self._trace)
                except Exception:
                    print(f"Invalid database: {filepath}")
            else:
                print(f"Invalid file: {filepath}")
        self.trace_line.focusNextChild()

    def get_trace_from_db(self) -> Optional[str]:
        """
        Show a custom Dialog to select a trace from a list of traces
        retrieved via de qtracedb configuration.
        :return: optional name of the trace
        """
        dlg = QtWidgets.QDialog(parent=self)
        dlg.setWindowTitle('Qtrace-DB connection')
        dlg.setObjectName("Dialog")

        v_layout = QtWidgets.QVBoxLayout(dlg)
        combobox = QtWidgets.QComboBox(dlg)
        v_layout.addWidget(combobox)
        buttonbox = QtWidgets.QDialogButtonBox(dlg)
        buttonbox.setOrientation(QtCore.Qt.Horizontal)
        buttonbox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel | QtWidgets.QDialogButtonBox.Ok)
        v_layout.addWidget(buttonbox)

        # Fill list of traces
        combobox.addItems(self._dbm.list_traces())

        buttonbox.accepted.connect(dlg.accept)
        buttonbox.rejected.connect(dlg.reject)

        dlg.exec()

        if dlg.result():
            return combobox.currentText()
        else:
            return None



# third-party modules
from PyQt5 import QtWidgets

# qsynthesis deps
from qsynthesis.plugin.dependencies import ida_kernwin, ida_bytes, ida_ua, ida_lines
from qsynthesis.types import Addr


POPUP_PATH = "QSynthesis/"


class SynthetizeFromHere(ida_kernwin.action_handler_t):
    """
    Popup action that will be shown in the context-menu of the IDA-View (right-click).
    This action just takes the current line address and put it in the 'from' field
    of the Qsynthesis view.
    """

    def __init__(self, widget: 'SynthesizerView'):
        """
        Constructor. Take the QSynthesis main view as parameter.
        """
        ida_kernwin.action_handler_t.__init__(self)
        self.action_id = "Qsynthesis:from-here"
        self.text = "Synthesize from here"
        self.shortcut = "Ctrl+Shift+A"
        self.tooltip = "Start synthesizing from the current address"
        self.icon = 127 # GREEN_LIGHT  # 84  # BLOCK_DESCENT_FRM
        self.widget = widget

    def set_text_widget(self, ea: Addr) -> None:
        """
        Set the address given in parameter in the 'from' field of
        the Qsynthesis view.

        :param ea: address to set
        :return: Nones
        """
        self.widget.from_line.setText(f"{ea:#x}")

    def activate(self, ctx) -> bool:
        """
        Activation callback. Retrieve current address and set the from
        field of the QSynthesis view
        """
        ea = ida_kernwin.get_screen_ea()
        ea = ida_bytes.get_item_head(ea)  # Make sure we are on the head of the instruction
        if ida_bytes.is_code(ida_bytes.get_flags(ea)):
            self.set_text_widget(ea)
            return True
        else:
            QtWidgets.QMessageBox.critical(self.widget, "Invalid byte", f"Current address: {ea:#x} is not code")
            return False

    def update(self, _) -> int:
        """ Overridden method called on hook update """
        return ida_kernwin.AST_ENABLE_ALWAYS

    def register(self) -> bool:
        """
        Register the popup_action to the 'IDA View-A'.

        :return: True if the registration succeeded
        """
        idaview = ida_kernwin.find_widget("IDA View-A")
        if idaview is None:
            print("Can't find IDA View to attach action")
            return False
        else:
            action = ida_kernwin.action_desc_t(self.action_id, self.text, self, self.shortcut, self.tooltip, self.icon)
            res = ida_kernwin.register_action(action)
            res2 = ida_kernwin.attach_action_to_popup(idaview, None, self.action_id, POPUP_PATH)
            return res & res2

    def unregister(self) -> bool:
        """
        Remove popup action from the context menu of the view

        :return: True if unregistration succeeded
        """
        res = ida_kernwin.unregister_action(self.action_id)
        idaview = ida_kernwin.find_widget("IDA View-A")
        if idaview is None:
            print("Can't find IDA View to detach action")
            return False
        else:
            res2 = ida_kernwin.detach_action_from_popup(idaview, self.action_id)
        return res & res2


class SynthetizeToHere(SynthetizeFromHere):
    """
    Popup action that will be shown in the context-menu of the IDA-View (right-click).
    This action just takes the current line address and put it in the 'To' field
    of the Qsynthesis view.
    """

    def __init__(self, widget: 'SynthesizerView'):
        """
        Constructor. Take the QSynthesis main view as parameter.
        """
        super(SynthetizeToHere, self).__init__(widget)
        self.action_id = "Qsynthesis:to-here"
        self.text = "Synthesize up to here (included)"
        self.shortcut = "Ctrl+Shift+Z"
        self.tooltip = "Start synthesizing up to the current address (included)"
        self.icon = 120 # BREAKPOINT_PLUS

    def set_text_widget(self, ea: Addr) -> None:
        """
        Set the address given in parameter in the 'to' field of the Qsynthesis view.

        :param ea: address to set
        :return: None
        """
        self.widget.to_line.setText(f"{ea:#x}")


class SynthetizeOperand(SynthetizeFromHere):
    """
    Popup action that will be shown in the context-menu of the IDA-View
    and retrieve the current operand being selected. At the moment only
    full operand can be selected (e.g mov rax, [rbp + 4], 'rbp cannot
    be selected separately)
    """

    def __init__(self, widget: 'SynthesizerView'):
        """
        Constructor. Take the QSynthesis main view as parameter.
        """
        super(SynthetizeOperand, self).__init__(widget)
        self.action_id = "Qsynthesis:operand"
        self.text = "Synthesize operand"
        self.shortcut = "Ctrl+Shift+O"
        self.tooltip = "Synthesize the current operand at this address"
        self.icon = 12  # STAR_PLUS

    def activate(self, ctx) -> bool:
        """
        Upon clicking (or shortcut key typed) retrieve the current operand
        object and set various widget variable to be able to retrieve the
        information later on.

        :return: True if the operand is value and has successfully been retrieved
        """
        ea = ida_kernwin. get_screen_ea()
        ea = ida_bytes.get_item_head(ea)  # Make sure we are on the head of the instruction
        if ida_bytes.is_code(ida_bytes.get_flags(ea)):
            op_num = ida_kernwin.get_opnum()
            i = self.widget.get_instruction(ea)
            if i is None:
                QtWidgets.QMessageBox.critical(self.widget, "Invalid instruction", f"Cannot disassembl instruction at: {ea:#x}")
                return False
            op = i.operands[op_num]
            if op.is_register() or op.is_memory():
                self.widget.to_line.setText(f"{ea:#x}")
                self.widget.op_num = op_num
                self.widget.op_is_read = not op.is_written()

                # Switch to operand mode and set text
                self.widget.switch_to_target_operand()
                text = ida_lines.tag_remove(ida_ua.print_operand(ea, op_num))
                self.widget.operand_label.setText(f"{op_num}:{text} ({'write' if op.is_written() else 'read'})")
                return True
            else:
                QtWidgets.QMessageBox.critical(self.widget, "Invalid operand", f"Invalid operand type: {op.type}\nCannot synthesize such type")
                return False
        else:
            QtWidgets.QMessageBox.critical(self.widget, "Invalid byte", f"Current address: {ea:#x} is not code")
            return False



from triton import AST_NODE
from typing import TypeVar, Union, Tuple, List, Generator, Dict, Optional, Iterable, Any, Iterator
from enum import IntEnum

# Utils types
BitSize = int
"""Integer representing a size in bits"""

ByteSize = int
"""Integer representing a size in bytes"""

Addr = int
"""Integer representing an address"""

# Triton Types
AstNode = TypeVar("AstNode")
"""`AstNode <https://triton.quarkslab.com/documentation/doxygen/py_AstNode_page.html>`_ SMT logic formula as returned by Triton"""

AstType = IntEnum("AstType", {k: v for k, v in AST_NODE.__dict__.items() if isinstance(v, int)})
"""Enum of all AstNode possible types as defined by the `AST_NODE <https://triton.quarkslab.com/documentation/doxygen/py_AST_NODE_page.html>`_ object of Triton"""

Register = TypeVar("Register")
""""`Register <https://triton.quarkslab.com/documentation/doxygen/py_Register_page.html>`_ object as returned by Triton"""

SymbolicExpression = TypeVar("SymbolicExpression")
"""`SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_ object as used by Triton"""

SymbolicVariable = TypeVar("SymbolicVariable")
"""`SymbolicVariable <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicVariable_page.html>`_ object as generated by Triton"""

# TritonAst & Grammar Types
Expr = AstNode
"""Alias for :py:obj:`AstNode`, to abstract to Triton type"""

SymVarMap = Dict[str, SymbolicVariable]
"""Dictionnary of symbolic variables names to theirs Triton object instance"""


# Synthesis types
Hash = Union[bytes, int, Tuple[int]]
"""Hash abstract type used to performed lookup in tables. Given the HashType of the
table a Hash value might be hashed before being used to perform the lookup."""

Char = str
"""String of size one. Used to represent normalized variables names"""

Input = Dict[Char, int]
"""An Input is a valuation of all inputs variables of an expression"""

Output = int
"""Value resulting of evaluating an Input on a given expression"""

IOPair = Tuple[Input, Output]
"""Represent a given input valuation and the resulting output integer"""

IOVector = List[IOPair]
"""List of input-outputs pairs"""



# Disclaimer:
# This file is a partial rip of the processor module of idacore
# it is meant to be replaced by idacore in a near future

# built-in modules
from enum import IntEnum

# third-party modules
from triton import ARCH

# qsynthesis modules
from qsynthesis.plugin.dependencies import ida_idp, ida_idaapi
from qsynthesis.plugin.dependencies import ArchX86, ArchX64, ArchARM, ArchARM64


class ProcessorType(IntEnum):
    """
    Enum of IDA Processor supported
    """
    UNKNOWN = -1
    PLFM_386 = ida_idp.PLFM_386       # Intel 80x86
    PLFM_ARM = ida_idp.PLFM_ARM       # Advanced RISC Machines


SUPPORTED_PROC = [ProcessorType.PLFM_386.value, ProcessorType.PLFM_ARM]


class ArchMode(IntEnum):
    """
    Enum of supported modes for each architectures
    """
    MODE32 = 0
    MODE64 = 1


class Processor:
    """
    Small idp wrapper to represent the current processor.
    """
    name = ida_idp.get_idp_name()
    id = ida_idp.ph_get_id()
    flag = ida_idp.ph_get_flag()
    type = ProcessorType.UNKNOWN if id not in SUPPORTED_PROC else ProcessorType(id)
    mode = ArchMode.MODE64 if ida_idaapi.get_inf_structure().is_64bit() else ArchMode.MODE32


def processor_to_triton_arch() -> ARCH:
    """
    Get the current IDB processor as a Triton ARCH object.

    :return: Triton ARCH from IDA processor type
    """
    if Processor.type == ProcessorType.PLFM_386:
        return ARCH.X86_64 if Processor.mode == ArchMode.MODE64 else ARCH.X86
    elif Processor.type == ProcessorType.PLFM_ARM:
        return ARCH.AARCH64 if Processor.mode == ArchMode.MODE64 else ARCH.ARM32
    else:
        assert False


def processor_to_arch():
    """
    Get the current IDB processor as a Qtrace-DB Arch object

    :return: Qtrace-DB Arch object from IDA processor type
    """
    if Processor.type == ProcessorType.PLFM_386:
        return ArchX64 if Processor.mode == ArchMode.MODE64 else ArchX86
    elif Processor.type == ProcessorType.PLFM_ARM:
        return ArchARM64 if Processor.mode == ArchMode.MODE64 else ArchARM
    else:
        assert False



from typing import List, Optional

import capstone
from capstone import x86_const, arm_const, arm64_const, CS_AC_READ, CS_AC_WRITE
from enum import IntEnum
from collections import namedtuple

Reg = namedtuple("Reg", "name")


class Opnd:
    def __init__(self, cs_op, arch):
        self._cs_op = cs_op
        self._arch = arch

    @property
    def type(self):
        return self._cs_op.type

    def is_read(self) -> bool:
        return bool(self._cs_op.access & CS_AC_READ)

    def is_written(self) -> bool:
        return bool(self._cs_op.access & CS_AC_WRITE)

    def is_register(self) -> bool:
        return self.type == self._arch.optypes.REG

    def is_memory(self) -> bool:
        return self.type == self._arch.optypes.MEM


class Instr:
    def __init__(self, cs_ins, arch):
        self._cs_ins = cs_ins
        self._arch = arch

    @property
    def bytes(self):
        return bytes(self._cs_ins.bytes)

    def __str__(self):
        return self._cs_ins.mnemonic + " " + self._cs_ins.op_str

    @property
    def operands(self):
        return [Opnd(x, self._arch) for x in self._cs_ins.operands]


class Arch:
    _CSD = None

    def disasm(cls, asm: bytes, addr: int) -> List[Instr]:
        cls._CSD.detail = True
        return [Instr(x, cls) for x in cls._CSD.disasm(asm, addr)]

    def disasm_one(cls, asm: bytes, addr: int) -> Instr:
        r = cls.disasm(asm, addr)
        return r[0] if r else None



class _ArchX86(Arch):
    NAME = "x86"
    INS_PTR = Reg('eip')
    STK_PTR = Reg('esp')
    _CSD = capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_32)
    nop_instruction = b"\x90"

    class optypes(IntEnum):
        INVALID = x86_const.X86_OP_INVALID
        IMM = x86_const.X86_OP_IMM
        REG = x86_const.X86_OP_REG
        MEM = x86_const.X86_OP_MEM


class _ArchX64(_ArchX86):
    NAME = "x86_64"
    INS_PTR = Reg('rip')
    STK_PTR = Reg('rsp')
    _CSD = capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_64)



class _ArchARM(Arch):
    NAME = "ARM"
    INS_PTR = Reg('pc')
    STK_PTR = Reg('sp')
    _CSD = capstone.Cs(capstone.CS_ARCH_ARM, capstone.CS_MODE_ARM)
    nop_instruction = b"\x00\xf0\x20\xe3"

    class optypes(IntEnum):
        INVALID = arm_const.ARM_OP_INVALID
        REG = arm_const.ARM_OP_REG
        IMM = arm_const.ARM_OP_IMM
        MEM = arm_const.ARM_OP_MEM
        FP = arm_const.ARM_OP_FP
        CIMM = arm_const.ARM_OP_CIMM
        PIMM = arm_const.ARM_OP_PIMM
        SETEND = arm_const.ARM_OP_SETEND
        SYSREG = arm_const.ARM_OP_SYSREG


class _ArchARM64(Arch):
    NAME = "AARCH64"
    INS_PTR = Reg('x28')
    STK_PTR = Reg('sp')
    _CSD = capstone.Cs(capstone.CS_ARCH_ARM64, capstone.CS_MODE_ARM)
    nop_instruction = b"\x1f\x20\x03\xd5"

    class optypes(IntEnum):
        INVALID = arm64_const.ARM64_OP_INVALID
        REG = arm64_const.ARM64_OP_REG
        IMM = arm64_const.ARM64_OP_IMM
        MEM = arm64_const.ARM64_OP_MEM
        FP = arm64_const.ARM64_OP_FP
        CIMM = arm64_const.ARM64_OP_CIMM
        REG_MRS = arm64_const.ARM64_OP_REG_MRS
        REG_MSR = arm64_const.ARM64_OP_REG_MSR
        PSTATE = arm64_const.ARM64_OP_PSTATE
        SYS = arm64_const.ARM64_OP_SYS
        PREFETCH = arm64_const.ARM64_OP_PREFETCH
        BARRIER = arm64_const.ARM64_OP_BARRIER


ArchX86 = _ArchX86()
ArchX64 = _ArchX64()
ArchARM = _ArchARM()
ArchARM64 = _ArchARM64()


class ArchsManager:
    @staticmethod
    def get_supported_regs(arch: Arch) -> List[Reg]:
        if isinstance(arch, _ArchX64):
            return [Reg('RAX'), Reg('RBX'), Reg('RCX'), Reg('RDX'), Reg('RDI'), Reg('RSI'), Reg('RBP'), Reg('RSP'),
                    Reg('RIP'), Reg('EFLAGS'), Reg('R8'), Reg('R9'), Reg('R10'), Reg('R11'), Reg('R12'), Reg('R13'),
                    Reg('R14'), Reg('R15')]
        elif isinstance(arch, _ArchX86):
            return [Reg('EAX'), Reg('EBX'), Reg('ECX'), Reg('EDX'), Reg('EDI'), Reg('ESI'), Reg('EBP'), Reg('ESP'), Reg('EIP'), Reg('EFLAGS')]
        elif isinstance(arch, _ArchARM):
            return [Reg('R0'), Reg('R1'), Reg('R2'), Reg('R3'), Reg('R4'), Reg('R5'), Reg('R6'), Reg('R7'), Reg('R8'),
                    Reg('R9'), Reg('R10'), Reg('R11'), Reg('R12'), Reg('R13'), Reg('R14'), Reg('R15'), Reg('CPSR')]
        elif isinstance(arch, _ArchARM64):
            return [Reg('X0'), Reg('X1'), Reg('X2'), Reg('X3'), Reg('X4'), Reg('X5'), Reg('X6'), Reg('X7'), Reg('X8'),
                    Reg('X9'), Reg('X10'), Reg('X11'), Reg('X12'), Reg('X13'), Reg('X14'), Reg('X15'), Reg('X16'),
                    Reg('X17'), Reg('X18'), Reg('X19'), Reg('X20'), Reg('X21'), Reg('X22'), Reg('R23'), Reg('X24'),
                    Reg('X25'), Reg('X26'), Reg('X27'), Reg('X28'), Reg('X29'), Reg('X30')]
        else:
            assert False



from qsynthesis.grammar.ops import BvOp
import pydffi
import array

# First, declare an FFI context
CODE = '''
#include <stdio.h>
#include <stdint.h> 
uint64_t add(uint64_t a, uint64_t b) { return a+b; }
uint64_t and(uint64_t a, uint64_t b) { return a&b; }
uint64_t or(uint64_t a, uint64_t b) { return a|b; }
uint64_t xor(uint64_t a, uint64_t b) { return a^b; }
uint64_t sub(uint64_t a, uint64_t b) { return a-b; }
uint64_t mul(uint64_t a, uint64_t b) { return a*b; }
uint64_t udiv(uint64_t a, uint64_t b) { return a/b; }
uint64_t usub(uint64_t a) { return -a; }
uint64_t invert(uint64_t a) { return ~a; }
uint64_t to_uint(int64_t a) { return (uint64_t) a; }
uint64_t urem(uint64_t a, uint64_t b) { return  b==0? a: a % b ; }
uint64_t ashr(uint64_t a, uint64_t b) { int bitsz = sizeof(uint64_t)*8; return (b >= bitsz)? ((int64_t) a) >> bitsz-1: ((int64_t) a) >> b; }
uint64_t sle(int64_t a, int64_t b) { return a <= b ; }
uint64_t slt(int64_t a, int64_t b) { return a < b ; }
uint64_t sge(int64_t a, int64_t b) { return a >= b ; }
uint64_t sgt(int64_t a, int64_t b) { return a > b ; }
uint64_t lshift(uint64_t a, uint64_t b) { return (b >= sizeof(uint64_t)*8)? 0: a << b; }
uint64_t rshift(uint64_t a, uint64_t b) { return (b >= sizeof(uint64_t)*8)? 0: a >> b; }
uint64_t rol(uint64_t i, uint64_t n) { int bitsz = 8*sizeof(n); return (n << (i%bitsz)) | (n >> (8*sizeof(n) - (i%bitsz))); }
uint64_t ror(uint64_t i, uint64_t n) { int bitsz = 8*sizeof(n); return (n >> (i%bitsz))|(n << (8*sizeof(n) - (i%bitsz))); }
uint64_t mod(uint64_t a, uint64_t b) { return a % b; }


void add_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] + b[i]; } }
void and_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] & b[i]; } }
void or_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] | b[i]; } }
void xor_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] ^ b[i]; } }
void sub_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] - b[i]; } }
void mul_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] * b[i]; } }
void udiv_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] / b[i]; } }
void usub_arr(uint64_t* dst, uint64_t* a, size_t n) { for(int i=0; i < n; i ++) { dst[i] = -a[i]; } }
void invert_arr(uint64_t* dst, uint64_t* a, size_t n) { for(int i=0; i < n; i ++) { dst[i] = ~a[i]; } }
void urem_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = b[i]==0? a[i]: a[i] % b[i] ; } }
void ashr_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { int bitsz = sizeof(uint64_t)*8; for(int i=0; i < n; i ++) { dst[i] = (b[i] >= bitsz)? ((int64_t) a[i]) >> bitsz-1: ((int64_t) a[i]) >> b[i]; } }
void sle_arr(uint64_t* dst, int64_t* a, int64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] <= b[i] ; } }
void slt_arr(uint64_t* dst, int64_t* a, int64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] < b[i] ; } }
void sge_arr(uint64_t* dst, int64_t* a, int64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] >= b[i] ; } }
void sgt_arr(uint64_t* dst, int64_t* a, int64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] > b[i] ; } }
void lshift_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = (b[i] >= sizeof(uint64_t)*8)? 0: a[i] << b[i]; } }
void rshift_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = (b[i] >= sizeof(uint64_t)*8)? 0: a[i] >> b[i]; } }
void rol_arr(uint64_t* dst, uint64_t* i, uint64_t* n, size_t sz) { for(int j=0; j < sz; j ++) { int bsz = 8*sizeof(n[j]); dst[j] = (n[j] << (i[j]%bsz)) | (n[j] >> (bsz - (i[j]%bsz))); } }
void ror_arr(uint64_t* dst, uint64_t* i, uint64_t* n, size_t sz) { for(int j=0; j < sz; j ++) { int bsz = 8*sizeof(n[j]); dst[j] = (n[j] >> (i[j]%bsz)) | (n[j] << (bsz - (i[j]%bsz))); } }
void mod_arr(uint64_t* dst, uint64_t* a, uint64_t* b, size_t n) { for(int i=0; i < n; i ++) { dst[i] = a[i] % b[i] ; } }
'''

ffi_ctx = pydffi.FFI(lazyJITWrappers=False)


def make_compilation_unit(bitsize: int):
    global ffi_ctx
    good_code = CODE.replace("64", str(bitsize))
    return ffi_ctx.compile(good_code)


def _get_func_name(op: 'Operator') -> str:
    return {
        BvOp.NOT: "invert",
        BvOp.AND: "and",
        BvOp.OR: "or",
        BvOp.XOR: "xor",
        BvOp.NEG: "usub",
        BvOp.ADD: "add",
        BvOp.MUL: "mul",
        BvOp.SUB: "sub",
        BvOp.SHL: "lshift",
        BvOp.LSHR: "rshift",
        BvOp.ROL: "rol",
        BvOp.ROR: "ror",
        BvOp.ASHR: "ashr",
    }[op.id]


def get_op_eval(cu, op):
    return getattr(cu.funcs, _get_func_name(op))


def get_op_eval_array(cu, op):
    return getattr(cu.funcs, f"{_get_func_name(op)}_arr")


def get_native_array_type(size_t, size):
    global ffi_ctx
    typ = {
        8: ffi_ctx.UInt8Ty,
        16: ffi_ctx.UInt16Ty,
        32: ffi_ctx.UInt32Ty,
        64: ffi_ctx.UInt64Ty
    }
    return ffi_ctx.arrayType(typ[size_t], size)


def init_array_cst(ffi_array, val, size, size_t) -> None:
    map = {8: 'B', 16: 'H', 32: 'I', 64: 'Q'}
    pydffi.view_as_bytes(ffi_array)[:] = array.array(map[size_t], [val]*size).tobytes()



# built-in modules
from __future__ import annotations
import logging

# qsynthesis deps
from qsynthesis.tables.base import InputOutputOracle
from qsynthesis.tritonast import TritonAst
from qsynthesis.types import Input, List, Optional, Tuple, Union, Output

logger = logging.getLogger("qsynthesis")


class SynthesizerBase:
    """
    Base Synthesizer that provides base function for children classes.
    It provides function for a given TritonAst to evaluate it against
    oracle inputs and to perform the lookup in order to know
    if a shorter expression exists.
    """

    def __init__(self, ltms: Union[InputOutputOracle, List[InputOutputOracle]], only_first_match: bool = False, learning_new_exprs: bool = False):
        """
        Constructor that takes one or multiple oracles as input.

        :param ltms: Single oracle or a list
        :param only_first_match: boolean that stop interating over tables as soon as the lookup is successfull for one
        :param learning_new_exprs: boolean that enables improving the current table if if a synthesized entry appears
                                   to be bigger than the one submitted
        """
        self._ltms = [ltms] if isinstance(ltms, InputOutputOracle) else ltms
        self.only_first_match = only_first_match
        self.learning_enabled = learning_new_exprs

        # Caches use internally
        self.expr_cache = {}  # Dict expr_str -> synth_expr
        self.eval_cache = {}  # Dict associating a the hash of an AST to a map of inputs -> output

        # Stats
        self.call_to_eval = 0
        self.cache_hit = 0
        self.eval_count = 0

    def synthesize(self, ioast: TritonAst, check_sem: bool = False) -> Tuple[TritonAst, bool]:
        """
        Abstract function that synthesize the given TritonAst, into a smaller if it exists.
        The implementation of this function is delegated to children classes.

        :param ioast: TritonAst object to synthesize
        :param check_sem: boolean one whether to check the semantic equivalence of expression
                          before substituting them. That ensure soundness of the synthesis
        :returns: tuple with new TritonAst and whether some replacement took place or not

        .. warning:: Activating the `check_sem` parameter implies a strong overhead
                     on the synthesis as SMT queries are being performed for any candidates
        """
        raise NotImplementedError("Should be implemented in children class")

    def try_synthesis_lookup(self, cur_ast: TritonAst, check_sem: bool = False) -> Optional[TritonAst]:
        """
        Performs a direct synthesis lookup. And returns an optional TritonAst if it the
        I/O evaluation has been found in one of the tables. Unlike :meth:`SynthesizerBase.synthesize`
        which can go down the AST to try simplifying sub-AST here only the root node is
        attempted to be synthesized.

        :param cur_ast: TritonAst to synthesize
        :param check_sem: boolean on whether to check the semantic equivalence of expression
                          before performing substituting.
        :returns: optional TritonAst if the the AST has been synthesized
        """

        # Try all lookup-tables and pick the shortest expression
        best_len = 0
        best_expr = None
        vars_e = cur_ast.symvars
        size_e = cur_ast.node_count

        if cur_ast.hash in self.expr_cache:
            self.cache_hit += 1
            logger.debug("expression cache found !")
            return self.expr_cache[cur_ast.hash]

        for table in self._ltms:
            # skip tables not having enough variables or the wrong bitsize
            if table.var_number < len(vars_e) or table.bitsize != cur_ast.size:
                continue

            synth_ast = self.run_direct_synthesis(table, cur_ast)
            if synth_ast is None:
                continue

            # Check if we got something better
            e_node_cnt = synth_ast.node_count
            if (best_expr is None and e_node_cnt < size_e) or (e_node_cnt < best_len):
                logger.debug(f"[base] candidate expr accepted: current:{size_e} candidate:{e_node_cnt} (best:{best_len})  => {synth_ast.pp_str}")
                best_len, best_expr = e_node_cnt, synth_ast
                if self.only_first_match:
                    break
            else:
                pass

            if best_len == 1:  # It cannot go better that this, so we can stop
                break

        if best_expr is not None:
            self.expr_cache[cur_ast.hash] = best_expr
            if check_sem:
                if cur_ast.is_semantically_equal(best_expr):
                    logger.info("Expressions are semantically equal")
                else:
                    logger.error("Expressions are semantically different (return None)!!")
                    return None
            return best_expr

    def run_direct_synthesis(self, ltm: InputOutputOracle, cur_ast: TritonAst) -> Optional['TritonAst']:
        """
        Evaluate `cur_ast` on inputs provided by `ltm` the oracle which provide an
        output vector then used to perform the lookup in the database. If an entry is found
        it is returned.

        :param ltm: InputOutputOracle object in which to perform the query
        :param cur_ast: TritonAst object to synthesize
        :returns: optional TritonAst if the an entry was found
        """

        # Evaluate node on LTMs inputs
        outputs = [self.eval_ast(cur_ast, i) for i in ltm.inputs]

        if len(set(outputs)) == 1:  # If all outputs are equal then we consider this as a constant expression
            logger.debug(f"[base] Found constant expression in {ltm.name}: {cur_ast.pp_str} ===> {outputs[0]}")
            return cur_ast.mk_constant(outputs[0], cur_ast.size)
        else:
            # Lookup expression with same outputs in the LT
            lk_expr = ltm.lookup(outputs, cur_ast, use_cache=False)
            if lk_expr is None:
                return None
            else:
                if lk_expr.node_count > cur_ast.node_count:
                    logger.debug(f"[base] synthesized bigger expression ({lk_expr.pp_str}) than given ({cur_ast.pp_str})")
                    if ltm.is_writable and self.learning_enabled:
                        h = ltm.hash(outputs)
                        s = cur_ast.to_normalized_str()
                        logger.info(f"[base] expression {s} added to {ltm.name}")
                        ltm.add_entry(h, s)
                return lk_expr

    def eval_ast(self, ioast: TritonAst, input: Input) -> Output:
        """
        Run evaluation of the TritonAst `ioast` on the given Input (valuation for all vars).
        The result is an Output (integer)

        :param ioast: TritonAst to evaluate
        :type ioast: TritonAst
        :param input: Input on which to evaluate the AST. All variables of ioast must
                      be defined in input
        :type input: :py:obj:`qsynthesis.types.Input`
        :returns: Output which is the result of evaluation (made by Triton)
        :rtype: :py:obj:`qsynthesis.types.Output`
        """
        self.call_to_eval += 1
        tup_inputs = tuple(input.items())

        ast_hash = ioast.hash
        if ast_hash in self.eval_cache:
            if tup_inputs in self.eval_cache[ast_hash]:
                return self.eval_cache[ast_hash][tup_inputs]
        else:
            self.eval_cache[ast_hash] = {}

        output = ioast.eval_oracle(input)
        self.eval_count += 1
        self.eval_cache[ast_hash][tup_inputs] = output
        return output
